### Main layer detection engine ###

import itertools
import heapq
import numpy as np
import scipy.spatial.distance as distance
import scipy.optimize
import functools
from operator import attrgetter
from Bio import PDB
from .bundle import get_local_axis

def create_pymol_selection_from_socket_results(indices):
	"""create pymol-readable selection from socket data (converted to bundleDesc)
	input: list of helices, each helix represented as a tuple in format (first_residue_number, last_residue_number, chain_id)
	output: list of pymol-readable selections, each as a string
	"""

	selection = []

	for idx in indices:
		selection.append('(chain {} and resi {}-{})'.format(idx[2], idx[0], idx[1]))

	return selection

def extract_CA_coords(helices, model, verbose=False):
	"""get CA coordinates for rach residue in each helix
	input: helix list, each helix is list of its residues in format (chain, ('', resnum, ''))
	output: list of helices, each helix is list of CA coords (Biopython Vector class) of its residues
	"""

	print('Extracting CA coordinates...')
	helices_CA = []
	for helix in enumerate(helices):

		#ignore KeyError structures (i.e. those with HETATM)
		try:
			helix_CA = [ model[res[0]][res[1]]['CA'].get_vector() for res in helix[1] ]
		except KeyError:
			return KeyError

		if verbose:
			print('CA COORDS: HELIX ' + str(helix[0]))
			for h in helix_CA:
				print(h)
		helices_CA.append(helix_CA)

	return helices_CA

def convert_Vector_data_to_np_array(helices_CA_Vector):
	"""converts list of helices format from Biopython Vector to lists of 3 points
	input: list of helices, each helix is list of Vectors, each is coords for one residue
	output: list of helices, each helix is list of 3-element lists describing residue coords
	"""

	return [ np.array([ (v[0], v[1], v[2]) for v in helix_CA ]) for helix_CA in helices_CA_Vector ]

def select_minimal_angle_layer_set2(layers_sets, best_layer_nb=1):

	def check_layers_shape(layers_set):

		def calculate_plane_equation_from_points(points):

			def plane(x, y, params):
				a = params[0]
				b = params[1]
				c = params[2]
				z = a*x + b*y + c
				return z

			def error(params, points):
				result = 0
				for (x,y,z) in points:
					plane_z = plane(x, y, params)
					diff = abs(plane_z - z)
					result += diff**2
				return result

			def cross(a, b):
				return [a[1]*b[2] - a[2]*b[1],
						a[2]*b[0] - a[0]*b[2],
						a[0]*b[1] - a[1]*b[0]]

			fun = functools.partial(error, points=points)
			params0 = [0, 0, 0]
			res = scipy.optimize.minimize(fun, params0)

			return res

		def calculate_angle_between_planes(plane1, plane2, silent_warning=True):

			"""calculate angle between two planes defined with equation ax+bx+cx+d
			cos = |a1a2 + b1b2 + c1c2| / (sqrt(a1^2 + b1^2 + c1^2) * sqrt(a2^2 + b2^2 + c2^2))
			input: (plane1/2) plane equation in form of dicionary where keys correspond to coefficients of equation
			output: angle between planes in degrees (float)
			"""

			numerator  = abs(sum([ plane1[coef]*plane2[coef] for coef in 'abc' ]))
			len_plane1 = np.sqrt(sum([ np.power(plane1[coef], 2) for coef in 'abc' ]))
			len_plane2 = np.sqrt(sum([ np.power(plane2[coef], 2) for coef in 'abc' ]))

			cos_phi    = numerator / (len_plane1 * len_plane2)
			if (1+1e-8 > cos_phi > 1.0):
				cos_phi = 1.0
				if not silent_warning:
					print('Warning: calculated cosine slightly beyound function codomain(max 1e-8 beyound) - corrected to 1.0.')
			phi_deg    = np.rad2deg(np.arccos(cos_phi))

			return phi_deg

		layer_equations = []
		layer_angles    = []

		for layer in layers_set.iterlayer():

			points = [ tuple(point.coords) for point in layer ]

			plane_equation = calculate_plane_equation_from_points(points)
			layer_equations.append({'a':plane_equation.x[0], 'b':plane_equation.x[1], 'c':plane_equation.x[2]})

		for layer in range(len(layer_equations)-1):
			# calculate angle between two planes
			layer_angles.append(calculate_angle_between_planes(layer_equations[layer], layer_equations[layer+1]))

		layers_set.average_dist_angle = np.median(layer_angles) #FIXME np.mean(layer_angles)
		print('V2', layers_set.average_dist_angle)

		return layers_set

	# this will return same layer sets list but with layers with set avg angle attribute
	layer_set_angles = (list(map(check_layers_shape, layers_sets)))

	if best_layer_nb == 1:
		best_layer_set_angle = min(layer_set_angles, key=attrgetter('average_dist_angle'))
	elif best_layer_nb == 'rank':
		best_layer_set_angle = sorted(layer_set_angles, key=attrgetter('average_dist_angle'))
	else:
		best_layer_set_angle = heapq.nsmallest(best_layer_nb, layer_set_angles, key=attrgetter('average_dist_angle'))

	return best_layer_set_angle

def select_minimal_angle_layer_set(layers_sets, best_layer_nb=1):
	"""find layer set with minimal angle between layers
	input: (CA_layers_list) list of lists of layers, each layer represented as n-element list of 3-element lists of coords (float) of CAs
	n=number of helices, topmost list comprise of lists of layers generated by scanning with different methods (from top or bottom of the bundle)
	input: (layer_ids_list) list of lists of identifiers (int) of CAs used to define first layer, each identifier represents one helix
	topmost list elements correspond to the elements of topmost list from CA_layers_list
	output: (best_layer_set) list of layers with minimal angle between them in format identical to elements of topmost list of CA_layers_list
	output: (best_layer_ids) element of layer_ids_list corresponding to selected best_layer_set element
	"""

	def check_layers_shape(layers_set):
		"""compute average angle between layers in bundle
		for each layer find all possible planes and compute angles between planes in vertical groups
		result is averaged angle from list of average angles from groups of planes
		input: list of layers, each layer is n-element list of 3-element lists (coords), n=oligomerization state
		output: average angle between layers (float)
		"""

		def calculate_plane_equation_from_points(x, y, z):
			a = np.column_stack((x, y, z))
			return np.linalg.lstsq(a, np.ones_like(x), rcond=None)[0]

		def calculate_angle_between_planes(plane1, plane2, silent_warning=True):

			"""calculate angle between two planes defined with equation ax+bx+cx+d
			cos = |a1a2 + b1b2 + c1c2| / (sqrt(a1^2 + b1^2 + c1^2) * sqrt(a2^2 + b2^2 + c2^2))
			input: (plane1/2) plane equation in form of dicionary where keys correspond to coefficients of equation
			output: angle between planes in degrees (float)
			"""

			numerator  = abs(sum([ plane1[coef]*plane2[coef] for coef in 'abc' ]))
			len_plane1 = np.sqrt(sum([ np.power(plane1[coef], 2) for coef in 'abc' ]))
			len_plane2 = np.sqrt(sum([ np.power(plane2[coef], 2) for coef in 'abc' ]))

			cos_phi    = numerator / (len_plane1 * len_plane2)
			if (1+1e-8 > cos_phi > 1.0):
				cos_phi = 1.0
				if not silent_warning:
					print('Warning: calculated cosine slightly beyound function codomain(max 1e-8 beyound) - corrected to 1.0.')
			phi_deg    = np.rad2deg(np.arccos(cos_phi))

			return phi_deg

		layer_equations = []
		layer_angles    = []

		for layer in layers_set.iterlayer():

			x = [ point.coords[0] for point in layer ] # all x coords from layer
			y = [ point.coords[1] for point in layer ]
			z = [ point.coords[2] for point in layer ]

			plane_equation = calculate_plane_equation_from_points(x,y,z)
			layer_equations.append({'a':plane_equation[0], 'b':plane_equation[1], 'c':plane_equation[2]})

		for layer in range(len(layer_equations)-1):
			# calculate angle between two planes
			layer_angles.append(calculate_angle_between_planes(layer_equations[layer], layer_equations[layer+1]))

		layers_set.average_dist_angle = np.median(layer_angles)

		return layers_set

	# this will return same layer sets list but with layers with set avg angle attribute
	layer_set_angles = (list(map(check_layers_shape, layers_sets)))

	if best_layer_nb == 1:
		best_layer_set_angle = min(layer_set_angles, key=attrgetter('average_dist_angle'))
	elif best_layer_nb == 'rank':
		best_layer_set_angle = sorted(layer_set_angles, key=attrgetter('average_dist_angle'))
	else:
		best_layer_set_angle = heapq.nsmallest(best_layer_nb, layer_set_angles, key=attrgetter('average_dist_angle'))

	return best_layer_set_angle

def find_bundle_boundry_layer_from_all(boundry_layers, distance_threshold, search_layer_setting_num):

	min_distance_set = heapq.nsmallest(search_layer_setting_num, boundry_layers)

	return [ layer for layer in min_distance_set if layer.total_distance <= distance_threshold ]

def select_minimal_distance_layer_set(layers_sets):
	"""find set of layers with minimal distance between 2 helices (dimers only)
	input/output same as described in select_minimal_angle_layer_set function
	only difference is that this function search for minimal distance between helices, not angle between layers
	"""

	def check_layers_distances(layers_set):

		layer_distances = []

		for layer in layers_set.iterlayer():
			point1 = layer[0].coords
			point2 = layer[1].coords

			layer_distances.append(distance.euclidean(point1, point2))

		layers_set.average_dist_angle = np.mean(layer_distances)

		return layers_set

	layer_set_distances     = (list(map(check_layers_distances, layers_sets)))
	best_layer_set_distance = min(layer_set_distances, key=attrgetter('average_dist_angle'))

	return best_layer_set_distance

def convert_to_bundleClass_format(samcc_selection):
	"""converts samcc selection from this library to original format used by bundleClass
	input: samcc-ready selection, list of helices, each helix a list of format [ chain_id(string), start_residue(int), stop_residue(int) ]
	output: BundleClass-ready selection, list of helices, each helix a tuple of format ( start_residue(int), stop_residue(int), chain_id(string), antiparallel(bool) )
	"""

	input_helices = []
	for helix in samcc_selection:
		if helix[1] < helix[2]:
			input_helices.append((helix[1], helix[2], helix[0], False))
		else:
			input_helices.append((helix[2], helix[1], helix[0], True))

	return input_helices

def select_minimal_dist_to_plane_layer_set(layers_sets, best_layer_nb=1):

	def check_layers_dist_to_plane(layers_set):

		def calculate_plane_equation_from_points(x, y, z):
			a = np.column_stack((x, y, z))
			return np.linalg.lstsq(a, np.ones_like(x), rcond=None)[0]

		def calculate_point_to_plane_distance(plane_equation, point):

			# plane coefficients
			A = plane_equation[0]
			B = plane_equation[1]
			C = plane_equation[2]

			# point coords
			x = point.coords[0]
			y = point.coords[1]
			z = point.coords[2]

			numerator   = abs(A*x + B*y + C*z)
			denominator = np.sqrt(A**2 + B**2 + C**2)

			return numerator / denominator

		layer_distances = []

		for layer in layers_set.iterlayer():

			x = [ point.coords[0] for point in layer ] # all x coords from layer
			y = [ point.coords[1] for point in layer ]
			z = [ point.coords[2] for point in layer ]

			plane_equation  = calculate_plane_equation_from_points(x,y,z)
			points_in_layer = []
			for point in layer:
				points_in_layer.append(calculate_point_to_plane_distance(plane_equation, point))

			layer_distances.append(np.mean(points_in_layer))

		layers_set.average_dist_to_plane = np.mean(layer_distances)

		return layers_set

	layer_set_dist_to_plane = (list(map(check_layers_dist_to_plane, layers_sets)))

	if best_layer_nb == 1:
		best_layer_set_dist_to_plane = min(layer_set_dist_to_plane, key=attrgetter('average_dist_to_plane'))
	elif best_layer_nb == 'rank':
		best_layer_set_dist_to_plane = sorted(layer_set_dist_to_plane, key=attrgetter('average_dist_to_plane'))
	else:
		best_layer_set_dist_to_plane = heapq.nsmallest(best_layer_nb, layer_set_dist_to_plane, key=attrgetter('average_dist_to_plane'))

	return best_layer_set_dist_to_plane

def select_minimal_total_distance_layer_set(layers_sets, best_layer_nb=1):

	def check_layers_total_distances(layers_set):

		def calculate_total_distance(layer, neighbour_interactions):
			"""Calculate total distance between neighbouring points in layer
			neighbourhood is determined according to neighbour_interactions list
			"""

			total_distance = 0

			for axis_point1 in layer:
				for axis_point2 in layer:
					if (axis_point1.helix_id, axis_point2.helix_id) in neighbour_interactions:
						total_distance += distance.euclidean(axis_point1.CA_coords, axis_point2.CA_coords)

			return total_distance

		layers_all_distances = []
		for layer in layers_set.iterlayer():
			layers_all_distances.append(calculate_total_distance(layer, layers_set.neighbour_interactions))

		layers_set.average_layers_dist = np.mean(layers_all_distances)

		return layers_set

	layer_set_total_distances = list(map(check_layers_total_distances, layers_sets))

	if best_layer_nb == 1:
		best_layer_set_total_distances = min(layer_set_total_distances, key=attrgetter('average_layers_dist'))
	elif best_layer_nb == 'rank':
		best_layer_set_total_distances = sorted(layer_set_total_distances, key=attrgetter('average_layers_dist'))
	else:
		best_layer_set_total_distances = heapq.nsmallest(best_layer_nb, layer_set_total_distances, key=attrgetter('average_layers_dist'))

	return best_layer_set_total_distances

def get_layers_set_with_best_ranks(layers_sets):

	best_layer_set_angle = sorted(layers_sets, key=attrgetter('average_dist_angle'))
	best_layer_set_dist_to_plane = sorted(layers_sets, key=attrgetter('average_dist_to_plane'))
	best_layer_set_total_distances = sorted(layers_sets, key=attrgetter('average_layers_dist'))

	for pos, layer_set in enumerate(best_layer_set_angle):
		layer_set.ranks += pos
	for pos, layer_set in enumerate(best_layer_set_total_distances):
		layer_set.ranks += pos

	return min(layers_sets, key=attrgetter('ranks'))

def find_closest_CA_to_point(boundry_layers, helices_axis_all):

	def get_closest_CA_to_axis_point(point, helix_pts):

		helix_point = helix_pts[0]
		dst_best    = distance.euclidean(helix_pts[0].CA_coords, point.coords)
		for helix_pt in helix_pts:
			dst = distance.euclidean(helix_pt.CA_coords, point.coords)
			if dst < dst_best:
				helix_point = helix_pt
				dst_best    = dst
		return helix_point

	boundry_layers_CA = []

	for layer, helix_axis in zip(boundry_layers, helices_axis_all):
		layer_adjusted = []
		for point in layer:
			closest_CA_helix_point = get_closest_CA_to_axis_point(point, [helix_axis[point.point_id+i] for i in range(-2,3) if point.point_id+i>0])
			layer_adjusted.append(closest_CA_helix_point)
		layer.axis_points = layer_adjusted
		boundry_layers_CA.append(layer)

	return boundry_layers_CA



import sys
import math
import numpy as np
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB import StructureBuilder
from Bio.PDB import PDBIO
from Bio.PDB import vector_to_axis
from Bio.PDB import Vector

def get_local_axis(calphas):
	"""Get local axis"""

	unit_local_axis = {}
	origin = {}
	if len(calphas)>=4:
		for j in range(len(calphas)-3):
			vec12 = calphas[j+1] - calphas[j]
			vec23 = calphas[j+2] - calphas[j+1]
			vec34 = calphas[j+3] - calphas[j+2]
			dv13 = vec12 - vec23
			dv24 = vec23 - vec34
			cross_product = (dv13 ** dv24).normalized()
			unit_local_axis[j] = cross_product

			dmag = math.sqrt(dv13._ar[0]*dv13._ar[0] + dv13._ar[1]*dv13._ar[1] + dv13._ar[2]*dv13._ar[2])
			emag = math.sqrt(dv24._ar[0]*dv24._ar[0] + dv24._ar[1]*dv24._ar[1] + dv24._ar[2]*dv24._ar[2])
			dot_product = dv13 * dv24

			costheta = dot_product/(dmag*emag)
			costheta1 = 1.0 - costheta
			radmag = math.sqrt(dmag*emag) / (costheta1*2.0)
			dv13 = dv13 ** (1/dmag)
			rad = dv13**radmag
			origin[j] = calphas[j+1] - rad

			dv24 = dv24 ** (1/emag)
			rad = dv24**radmag
			origin[j+1] = calphas[j+2] - rad

	else:
		print("Helix is less than 4 residues and has no axis\n\n")
		sys.exit(-1)

	return list(origin.values())



### Helper functions ###

import networkx as nx
import itertools
import numpy as np
from math import factorial

def gen_expected_crick_angles(P, rep_len, start_ph1, ap=False):

	step = 360 / P
	if ap:
		sign=-1
	else:
		sign=1
	return [adj(start_ph1+(sign * i * float(step))) for i in range(rep_len)]

def adj2(ang):

	ang = adj(ang)
	if ang < -180:
		ang += 360
	return ang

def adj(ang):

	ang += 180.0
	if ang>360:
		c = int(ang / 360.0)
		return ang - (360*c) - 180.0
	else:
		return ang - 180.0

def adjustangle(angle):

	if abs(angle)>180 and angle > 0:
		angle = angle - 360.0
	elif abs(angle)>180 and angle < 0:
		angle += 360.0
	return angle

def diffangle(targetA, sourceA):

	a = targetA - sourceA
	a = (a + 180) % 360 - 180
	return a

def crick_to_pos(start_Cr_ang, exp_helix_crick):

	diff = [abs(adj(start_Cr_ang-i)) for i in exp_helix_crick]
	mindiff = min(diff)

	start_Cr_ang_pos = diff.index(mindiff)

	try:
		name = chr(97+start_Cr_ang_pos)
	except ValueError:
		name = '?'

	return start_Cr_ang_pos, name, mindiff

def calc_crick_ang_dev(twister, exp_helix_crick, firstpos, lastpos, force_start_Cr_ang_pos=None):

	if firstpos == lastpos:
		assert lastpos == None

	# define Crick angle of the starting position
	if force_start_Cr_ang_pos==None:
		if lastpos==None:
			firstpos = 0
		if type(twister) is list:
			start_Cr_ang = twister[firstpos]
		else:
			start_Cr_ang = twister.iloc[firstpos]['Cr_ang']
		start_Cr_ang_pos, name, _ = crick_to_pos(start_Cr_ang, exp_helix_crick)
	else:
		start_Cr_ang_pos = force_start_Cr_ang_pos

	cr = itertools.cycle(exp_helix_crick)

	# skip to the first repeat pos
	for n in range(start_Cr_ang_pos):
		next(cr)

	if lastpos==None:
		laspos = len(twister)
	if type(twister) is list:
		data = twister[firstpos:lastpos]
	else:
		data = twister['Cr_ang'].iloc[firstpos:lastpos]

	Cr_ang_dev = [adj2(c - next(cr)) for c in data]

	m,b = np.polyfit(range(len(Cr_ang_dev)), Cr_ang_dev, 1)

	return Cr_ang_dev, m, b

def window(seq, n=2):
	"""Returns a sliding window (of width n) over data from the iterable
	s -> (s0,s1,...s[n-1]), (s1,s2,...,sn), ...
	"""

	it = iter(seq)
	result = tuple(itertools.islice(it, n))
	if len(result) == n:
		yield result
	for elem in it:
		result = result[1:] + (elem,)
		yield result

def savitzky_golay(y, window_size, order, deriv=0, rate=1):
	"""Smooth (and optionally differentiate) data with a Savitzky-Golay filter.
	The Savitzky-Golay filter removes high frequency noise from data.
	It has the advantage of preserving the original shape and
	features of the signal better than other types of filtering
	approaches, such as moving averages techniques.
	Parameters
	----------
	y : array_like, shape (N,)
		the values of the time history of the signal.
	window_size : int
		the length of the window. Must be an odd integer number.
	order : int
		the order of the polynomial used in the filtering.
		Must be less then `window_size` - 1.
	deriv: int
		the order of the derivative to compute (default = 0 means only smoothing)
	Returns
	-------
	ys : ndarray, shape (N)
		the smoothed signal (or it's n-th derivative).
	"""

	try:
		window_size = np.abs(np.int(window_size))
		order = np.abs(np.int(order))
	except ValueError as msg:
		raise ValueError("window_size and order have to be of type int")
	if window_size % 2 != 1 or window_size < 1:
		raise TypeError("window_size size must be a positive odd number")
	if window_size < order + 2:
		raise TypeError("window_size is too small for the polynomials order")
	order_range = range(order+1)
	half_window = (window_size -1) // 2
	# precompute coefficients
	b = np.mat([[k**i for i in order_range] for k in range(-half_window, half_window+1)])
	m = np.linalg.pinv(b).A[deriv] * rate**deriv * factorial(deriv)
	# pad the signal at the extremes with
	# values taken from the signal itself
	firstvals = y[0] - np.abs( y[1:half_window+1][::-1] - y[0] )
	lastvals = y[-1] + np.abs(y[-half_window-1:-1][::-1] - y[-1])
	y = np.concatenate((firstvals, y, lastvals))
	return np.convolve( m[::-1], y, mode='valid')

# SOCKET helper functions

def detect_helices_orientation(indices, orientation_data):

	"""
	Arguments

	indices: dict of bundle helices from Socket
	orientation_data: relative orientation of the helices from indices

	Returns

	a dict helix_id -> orientation

	this function uses a graph to convert relative orientations between the helices
	to parallel/anti-parallel labels for each helix

	"""

	G=nx.Graph()
	for o in orientation_data:
		G.add_edge(o[0], o[1], o=o[2])

	res = {}

	for g in nx.connected_component_subgraphs(G):

		first=True
		for e in list(nx.dfs_edges(g)):
			if first:
				g.node[e[0]]['o'] = 1
				first = False

			if g[e[0]][e[1]]['o']=='antiparallel':
				g.node[e[1]]['o'] = g.node[e[0]]['o']*-1
			else:
				g.node[e[1]]['o'] = g.node[e[0]]['o']

		for n in g.nodes():
				res[n] = g.node[n]['o']

	return res



import pandas as pd
import sys
import os
import math
import time
import glob
import itertools
import numpy as np
from Bio.PDB.PDBParser import PDBParser
import matplotlib.pyplot as plt
import seaborn as sns
if sys.version_info[0] == 3:
	from io import StringIO
else:
	import StringIO

np.seterr(all='raise')

def parse(fname, drop=[], silent=False):
	try:
		tfile = open(fname).read()

		seq = tfile[tfile.find("********* Sequence and heptad assignment *********"):tfile.find("********* Coiled-coil parameters per residue *********")]

		s=[]
		i=3
		seq = seq.split('\n')
		while i<len(seq):
			#print i, s
			s+= list(seq[i].strip().replace(' ', ''))
			i+=4

		s = s[1:-1]

		tfile = StringIO.StringIO(tfile[tfile.find("Res cc_phase"):tfile.find("Ave")].replace(" R ", "   "))
		df = pd.read_csv(tfile, error_bad_lines=False, delim_whitespace=True, header=0, skiprows=0, index_col=False, engine='c', dtype={"total_score": np.float64})

		#print fname

		df['seq'] = s

	except pd.io.common.EmptyDataError:
		if not silent: print('Error reading %s. Empty file?' % fname)
		return None
	except IOError:
		if not silent: print('Error reading %s. File does not exist.' % fname)
		return None

	else:
		for d in drop:
			df.drop(d, axis=1, inplace=True)

		# skip log files that contain nan values
		if df.isnull().values.any():
			if not silent: print('Error reading %s. File constains NaN.' % fname)
			return None
		#df['P'] = df['res/tur']/(1-(df['res/tur']*df.cc_dang/360.0))
		df['P'] = 3.627/(1-(3.627*df.cc_dang/360.0))
		try:
			df['A'] = np.degrees(np.arcsin(np.radians(df['cc_dang']) * df['cc_rad']) / df['a_ris'])
		except FloatingPointError:
			if not silent: print('Can''t calculate A %s' % fname)
			#return None

		return df

def average(df, fname="", allCrick=False):

	cr = df.Cr_ang.tolist()
	if not allCrick:
		cr = cr[0]

	fit = {'tw_P':df.P.mean(),\
		   'tw_p':df['res/tur'].mean(),\
		   'tw_w0':df.cc_dang.mean(),\
		   'tw_R0':df.cc_rad.mean(),\
		   'tw_R1':df.a_rad.mean(),\
		   'tw_d':df.a_ris.mean(),\
		   'tw_ccrise':df.cc_rise.mean(),\
		   'tw_ccpit':df.cc_pit.mean(),\
		   'tw_w1':df.a_dang.mean(),\
		   #'tw_ph1':float(df.iloc[[0]]['Cr_ang']),\
		   'tw_ph1':cr,\
		   }

	fit.update({'P_std':df.P.std(),\
		   'tw_p_std':df['res/tur'].std(),\
		   'tw_w0_std':df.cc_dang.std(),\
		   'tw_R0_std':df.cc_rad.std(),\
		   'tw_R1_std':df.a_rad.std(),\
		   'tw_d_std':df.a_ris.std(),\
		   'tw_ccrise_std':df.cc_rise.std(),\
		   })

	if fname!="":
		fit['id'] = fname

	return fit

def getrange(pdbfile):
	p = PDBParser(PERMISSIVE=1)
	s = p.get_structure("", pdbfile)
	chnames=[]
	pf, pt= 0, 1000000
	for c in s[0]: # iterate over chains

		cl = list(c) # list of all residues in a chain

		#if len(cl)<25: continue # skip short chains

		chnames.append(c.get_id())
		c = cl

		pos = 0

		assert (c[0].get_full_id()[3][0])==' ' # first residue is not an HET atom

		while c[pos].get_full_id()[3][0]==' ' and pos<len(c)-1:
			pos+=1

		pos-=1

		f,t = c[0].get_id()[1], c[pos].get_id()[1]

		#print f,t

		if f>pf: pf=f
		if t<pt: pt=t



	return pf, pt, ' '.join(chnames)


def run_twister_local(pdbfile, force=False):
	logfile = pdbfile.replace('.pdb', '.log')
	f, t, chnames = getrange(pdbfile)

	if os.path.isfile(logfile) and not force:
		return logfile

	os.system("~/apps/twister/runscript %s %s %s %s" % (pdbfile, f, t))
	return logfile

def run_twister_remote(pdbfile, force=False, f=None, t=None, chnames=""):
	logfile = pdbfile.replace('.pdb', '.log')
	resfile = pdbfile.replace('.pdb', '.res')
	#os.system('rm %s' % logfile)

	if f==None and t==None:
		f, t, chnames = getrange(pdbfile)
	else:
		if chnames=="":
			_, _, chnames = getrange(pdbfile)


	if os.path.isfile(logfile) and not force:
		return f,t
		#return logfile


	# edi
	#os.system('scp %s sdunin@edi01:/home/users/sdunin/apps/twister' % pdbfile)
	#cmd='ssh sdunin@edi01 "cd /home/users/sdunin/apps/twister/; ./runscript %s %s %s %s"' % (pdbfile, f, t, chnames)
	#print cmd
	#os.system(cmd)
	#os.system('scp sdunin@edi01:/home/users/sdunin/apps/twister/%s .' % logfile)

	os.system('scp %s sdh@hades.cent.uw.edu.pl:/home/sdh/apps/twister' % pdbfile)
	cmd='ssh sdh@hades.cent.uw.edu.pl "cd /home/sdh/apps/twister/; ./runscript %s %s %s %s"' % (pdbfile, f, t, chnames)
	print(cmd)
	os.system(cmd)
	os.system('scp sdh@hades.cent.uw.edu.pl:/home//sdh/apps/twister/%s .' % logfile)


	time.sleep(1)
	#return logfile
	# Edi 01
	#os.system('scp -q %s sdunin@edi01:/home/users/sdunin/apps/twister/' % pdbfile)
	#os.system('ssh sdunin@edi01 "cd /home/users/sdunin/apps/twister/; ./runscript %s"' % pdbfile)
	#os.system('scp -q sdunin@edi01:/home/users/sdunin/apps/twister/%s .' % logfile)
	#os.system('ssh sdunin@edi01 "cd /home/users/sdunin/apps/twister/; rm %s; rm %s; rm %s"' % (pdbfile, logfile, resfile))

	return f,t

def adj(ang):
	ang += 180.0
	if ang>360:
		c = int(ang / 360.0)
		return ang - (360*c) - 180.0
	else:
		return ang - 180.0

def adj2(ang):
	ang = adj(ang)
	if ang < -180:
		ang += 360
	return ang

def calc_crick_ang_dev(twister, exp_helix_crick, firstpos, lastpos, force_start_Cr_ang_pos=None):

	if firstpos == lastpos:
		assert lastpos == None


	# define Crick angle of the starting position
	if force_start_Cr_ang_pos==None:
		if lastpos==None:
			firstpos = 0
		if type(twister) is list:
			start_Cr_ang = twister[firstpos]
		else:
			start_Cr_ang = twister.iloc[firstpos]['Cr_ang']
		start_Cr_ang_pos, name, _ = crick_to_pos(start_Cr_ang, exp_helix_crick)
	else:
		start_Cr_ang_pos = force_start_Cr_ang_pos

	cr = itertools.cycle(exp_helix_crick)

	# skip to the first repeat pos
	for n in range(start_Cr_ang_pos):
		next(cr)

	if lastpos==None:
		laspos = len(twister)
	if type(twister) is list:
		data = twister[firstpos:lastpos]
	else:
		data = twister['Cr_ang'].iloc[firstpos:lastpos]

	Cr_ang_dev = [adj2(c - next(cr)) for c in data]

	m,b = np.polyfit(range(len(Cr_ang_dev)), Cr_ang_dev, 1)

	return Cr_ang_dev, m, b

def gen_expected_crick_angles(P, rep_len, start_ph1, ap=False):
	step = 360 / P
	if ap:
		sign=-1
	else:
		sign=1
	return [adj(start_ph1+(sign * i * float(step))) for i in range(rep_len)]

def crick_to_pos(start_Cr_ang, exp_helix_crick):

	diff = [abs(adj(start_Cr_ang-i)) for i in exp_helix_crick]
	mindiff = min(diff)

	start_Cr_ang_pos = diff.index(mindiff)

	try:
		name = chr(97+start_Cr_ang_pos)
	except ValueError:
		name = '?'

	return start_Cr_ang_pos, name, mindiff

	#print "region %s-%s, starting position %s (%s degrees)" % (firstpos, firstpos+fragmentlen, chr(97+start_Cr_ang_pos), exp_helix_crick[start_Cr_ang_pos])

	#print start_Cr_ang_pos
	#print chr(97+start_Cr_ang_pos)

	#print 'measured starting ph1', start_Cr_ang
	#print 'closest ideal ph1', exp_helix_crick[start_Cr_ang_pos]



def measure_axial_rotation(logname, P=3.5, replen=7, start_ph1=19.5):

	exp_ahelix_crick = gen_expected_crick_angles(P, replen, start_ph1)

	id = logname[:-4].split('/')[-1]
	df = parse(logname)
	if not type(df) is pd.DataFrame:
		return None, None, None

	#print df

	start_Cr_ang = df.Cr_ang.iloc[0]
	start_Cr_ang_pos, name, _ = crick_to_pos(start_Cr_ang, exp_ahelix_crick)
	crdev = calc_crick_ang_dev(df, exp_ahelix_crick, 0, len(df))[0]

	df['crdev'] = crdev

	return id, df, name



def measure_axial_rotations_in_dir(dir, P=3.5, replen=7, start_ph1=19.5):

	"""
	parses all twister log files in a directory and caluclates helix axial rotations
	"""
	res = {}

	for logname in glob.glob(dir+'/*.log'):
		id, df, name = measure_axial_rotation(logname, P=P, replen=replen, start_ph1=start_ph1)
		if not id==None:

			twister_std_crdev = np.std(df.crdev)
			twister_mean_crdev = np.mean(df.crdev)

			res[id] = {'twister_std_crdev': twister_std_crdev, 'twister_mean_crdev': twister_mean_crdev, 'heptad':name}

	df = pd.DataFrame.from_dict(res, orient='index')
	df['description'] = df.index

	return df


# plotting

def do_plots(dfs, colors):
	TICKFONTSIZE = 10
	AXESFONTSIZE = 15

	fig, axes = plt.subplots(nrows=1, sharey=False, sharex=False, figsize=(8, 5))
	for df, color in zip(dfs, colors):

		#print df

		#ax1 = df.crdev.rolling(window=5,center=False).mean()
		print(df.crdev.std())
		axes.plot(df.crdev.tolist(), c=color, linewidth=4.0)
		#axes.axhline(ls='--', color='0.5', y=df.crdev.mean(), linewidth=2.0)

		#for i in [17/5., 7/2., 25/7., 18/5., 11/3., 15/4., 19/5.]:
		#	axes.axhline(ls='--', color='0.5', y=i, linewidth=2.0)

		axes.set_xlabel(r'Position in the sequence', fontsize=AXESFONTSIZE)
		axes.set_ylabel(r'Helix axial rotation (degrees)', fontsize=AXESFONTSIZE)



		axes.set_ylim(-26, 26)
		#plt.yticks(np.arange(-26, 26, 2.0))

		for tick in axes.xaxis.get_major_ticks():
			tick.label.set_fontsize(TICKFONTSIZE)

		for tick in axes.yaxis.get_major_ticks():
			tick.label.set_fontsize(TICKFONTSIZE)
	plt.show()
	#plt.savefig('/Users/sdh/Desktop/twister.png')





if __name__ == "__main__":


	pdbfile = sys.argv[1]

	"""
	try:
		import renumber_PDB
		renumber_PDB.write_structure(pdbfile)
	except:
		pass
	"""

	run_twister_remote(pdbfile)

	logfile = pdbfile.replace('.pdb', '.log')
	#df = parse(logfile)
	id, df, name=measure_axial_rotation(logfile)


	do_plots([df], ['red'])



### Socket parser from localpdb - version from 5.03.20 ###

import re
import itertools

def parse_socket_output(filename, method='overlap'):
    """
    Parser of the Socket (J. Mol. Biol., 307 (5), 1427-1450) coiled coil assignment program
    :param filename: Socket output (short version) filename
    :param method: Assignment method
                    'heptads' - based on the heptad assignment
                    'knobs' - based on the knobs assignment
                    'overlap' - sum of the ranges of 'heptads' and 'knobs'
    :return: dict with the parsed coiled coils informations (if present) or 0 (if coiled coils absent)
    """
    if type(filename) is not str:
        return 0
    assert method in ['overlap', 'heptads', 'knobs']

    # Read all lines from file
    try:
        f = open(filename, 'r')
        lines = [line.rstrip('\n') for line in f.readlines()]
        f.close()
    except (OSError, FileNotFoundError):
        return 0

    # Get indices (numbers of first line in file) for each CC assignment
    start_indices = []
    for i in range(0, len(lines)):
        line = lines[i]
        if line.startswith('coiled coil') and line.endswith(':'):
            start_indices.append(i)
        if 'Finished' in line:
            start_indices.append(i)

    # Iterate over all CC assignments
    all_coils = {}
    relations = []
    for i in range(0, len(start_indices) - 1):  # Last one is 'Finished' line
        start = start_indices[i]  # Starting line for CC assignment
        stop = start_indices[i + 1]  # End line for CC assignment (next assignment starts or 'Finished')
        cc_id = 'cc_{}'.format(lines[start].split()[2].replace(':', ''))
        # Data
        coil_info = {'helix_ids': [], 'indices': {}, 'sequences': {}, 'heptads': {}, 'ambigous': False, 'relations': []}
        for k in range(start, stop):
            line = lines[k].rstrip()
            if line.startswith('assigning heptad to helix'):
                temp = line.split(' ')
                # Get start and end residues for helix
                try:
                    if temp[6][0] == '-':  # First residue is negative
                        if temp[6].count('-') == 3:  # Both residues are negative
                            start_res = -int(temp[6].split('-')[1])
                            end_res = -int(temp[6].split('-')[3].split(':')[0])
                        else:
                            start_res = -int(temp[6].split('-')[1])
                            end_res = int(temp[6].split('-')[2].split(':')[0])
                    else:
                        start_res = int(temp[6].split('-')[0])
                        end_res = int(temp[6].split('-')[1].split(':')[0])
                    helix_id = 'helix_{}'.format(int(temp[4]))
                    chain = line.split(':')[1]
                except ValueError:
                    coil_info['ambigous'] = True
                coil_info['helix_ids'].append(helix_id)
                seq_line = lines[k + 2]
                register_line = lines[k + 3]
                knobtype_line = lines[k + 5]
                seq = seq_line[9:]
                if len(seq) != (end_res - start_res + 1):
                    coil_info['ambigous'] = True
                register = register_line[9:]
                knobtype = knobtype_line[9:]
                if method == 'heptads':
                    left_marg = len(register) - len(register.lstrip())
                    right_marg = len(register) - len(register.rstrip())
                elif method == 'knobs':
                    left_marg = len(knobtype) - len(knobtype.lstrip('-'))
                    right_marg = len(knobtype) - len(knobtype.rstrip('-'))
                elif method == 'overlap':
                    left_marg1 = len(register) - len(register.lstrip())
                    left_marg2 = len(knobtype) - len(knobtype.lstrip('-'))
                    right_marg1 = len(register) - len(register.rstrip())
                    right_marg2 = len(knobtype) - len(knobtype.rstrip('-'))
                    left_marg = min(left_marg1, left_marg2)
                    right_marg = max(right_marg1, right_marg2)
                start_res = start_res + left_marg
                end_res = end_res - right_marg
                fseq = seq[left_marg:len(seq) - right_marg]
                register = register[left_marg:len(seq) - right_marg].replace(' ', '-')
                coil_info['heptads'][helix_id] = register
                coil_info['sequences'][helix_id] = fseq
                coil_info['indices'][helix_id] = {'start': int(start_res), 'end': int(end_res), 'chain': chain}
            if 'length max' in line and 'PRESENT' not in line and 'REPEATS' not in line:
                inf = re.findall('\((.*?)\)', line)
                data = inf[1].split(' ')
                oligomerization = data[1][0]
                orientation = data[0]
                coil_info['oligomerization'] = oligomerization
                coil_info['orientation'] = orientation
            if line.startswith('	angle between helices'):
                rel_data = line.split()
                first_helix, second_helix, orientation = 'helix_{}'.format(rel_data[3]), 'helix_{}'.format(rel_data[5]), rel_data[8]
                coil_info['relations'].append([first_helix, second_helix, orientation])
        all_coils[cc_id] = coil_info
    if not all_coils:
        return {}
    else:
        return all_coils



import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import itertools
import pandas as pd
import warnings
from Bio.PDB.StructureBuilder import StructureBuilder
from Bio.PDB import PDBIO
from Bio.PDB import PDBParser
from .layerClass import layerClass
from .residueClass import residueClass
from .chainClass import chainClass
from .pymol_draw_layers import save_layers_to_pymol
#plt.switch_backend('SVG')

class bundleClass():
	"""Class describing single bundle of the structure"""

	def __init__(self):
		self.layers = []
		self.pymol_selection = []
		self.helix_order = []
		self.helices_axis = []
		self.ppo = []

	def calc_bundleaxis(self):
		"""Calculate axis of entire bundle"""

		axis=[]

		for l in self.layers[1:-1]:
			res_in_layer = [r.O for r in l.res]
			C = np.sum(res_in_layer) ** (1.0/len(self.chains))
			for r in l.res:
				r.C = C
			axis.append(C)

		self.axis = [None]+axis+[None]

	def get_helicesaxis(self):
		"""Get axis of individual helices"""

		for c in self.chains:
			self.helices_axis.append(c.axis)

	def calc_pitch_angle(self):
		"""Calculate pitch angle"""

		for c in self.chains:
			c.calc_pitch_angle()

	def calc_crick(self):
		"""Calculates Crick angles for all chains in the bundle."""

		for c in self.chains:
			c.calc_crick()

	def calc_crickdev(self, P, REP, optimal_ph1=19.5, smooth=False):
		"""Calculates Crick's angle deviations for all chains in a bundle
		P - periodicity (e.g. 3.5 for 7/2)
		REP - repeat length (eg. 7 for 7/2)
		"""
		for c in self.chains:
			c.calc_crickdev(P, REP, optimal_ph1=optimal_ph1, smooth=smooth)

	def calc_periodicity(self, smooth=False, helix_p=3.63):
		"""Calculates periodicity for all chains in a bundle."""
		for c in self.chains:
			c.calc_periodicity(smooth=smooth, helix_p=helix_p)

	def calc_radius(self):
		"""Calculates radius for all chains in a bundle."""
		for c in self.chains:
			c.calc_radius()

	def calc_axialshift(self):
		"""Calculates axial shift for all chains in a bundle."""
		for c in self.chains:
			c.calc_axialshift()

	def assign_positions(self):
		"""Assign heptad positions using TWISTER algorithm."""
		for c in self.chains:
			c.assign_positions()

	def write_helical_axes(self, filename):
		"""Writes helical axes in PDB format."""

		sb = StructureBuilder()
		sb.init_structure('')
		sb.init_model('')

		for cpos, chain in enumerate(self.chains):
			sb.init_chain(str(cpos))
			sb.init_seg('')

			for pos, i in enumerate(chain.res[1:-1]):
				sb.init_residue('ALA',' ',pos,' ')
				sb.init_atom('CA', i.O._ar,  0, 0, " ", ' CA ', 1)

		io=PDBIO()
		io.set_structure(sb.structure)
		io.save(filename)

	def from_defdata(self, filename, chains, chains_names, chains_ap, chains_heptad):
		"""Reads bundle based on provided parameters.

			Arguments:

				filename: name of the PDB file
		"""

		parser=PDBParser()
		structure=parser.get_structure('temp', filename)

		assert len(set([len(i) for i in chains]))==1, "Provided chains are not of equal length! Aborting!"

		# read layers
		for layer in zip(*chains):
			l=layerClass([ residueClass(structure[0][c][r], c) for r, c in zip(layer, [x[0] for x in chains_names]) ])
			self.layers.append(l)

		# read chains
		self.chains = [list([]) for _ in range(len(chains_names))]

		for l in self.layers:
			for pos, r in enumerate(l.res):
				self.chains[pos].append(r)

		for pos, c in enumerate(self.chains):
			self.chains[pos] = chainClass(self.chains[pos], chains_ap[pos], chains_names[pos], chains_heptad[pos])
			self.chains[pos].calc_axis()

	def gendf(self, cols=['radius', 'crick', 'P', 'shift', 'crdev', 'positions_twister', 'p', 'positions_samcc']):
		"""Generates pandas df containg all bundle parameters."""

		index, data = [], []
		warned = []

		for chainpos, chain in enumerate(self.chains):
			for rpos in range(len(chain.radius)):

				index.append((rpos, chainpos))

				row_dict = {'res_name':chain.res[rpos].res.get_resname(),
							'res_number':chain.res[rpos].res.full_id[3][1],
							'chain_name':chain.res[rpos].chain_name,
							'chain_ap':chain.ap}

				for c in cols:
					try:
						row_dict[c] = getattr(chain, c)[rpos]
					except AttributeError:
						if not c in warned:
							warnings.warn("%s data missing" % c)
							warned.append(c)

				data.append(row_dict)

		index = pd.MultiIndex.from_tuples(index, names=['layer', 'chain'])

		return pd.DataFrame(data, index=index).sort_index()

	def plot(self, filename="", elements=[]):

		"""Plot data measured by the layer"""

		assert len(elements)>1
		assert type(elements)==list

		sns.set_style("ticks")

		TICKFONTSIZE = 15
		AXESFONTSIZE = 15

		fig, axes = plt.subplots(nrows=len(elements), sharey=False, sharex=False, figsize=(15, len(elements)*3))
		a=0

		plt.subplots_adjust(hspace=0.5)

		colgen = itertools.cycle('gbmr')

		for e in elements:
			for chain, color in zip(self.chains, colgen):

				if e=='Periodicity':
					# Plot bundle periodicity
					axes[a].plot(chain.P[1:-1], c=color, linewidth=1.0)
					#for i in [17/5., 7/2., 25/7., 18/5., 11/3., 15/4., 19/5.]:
					#	axes[0].axhline(ls='--', color='0.5', y=i, linewidth=1.0)
					#axes[a].axhline(ls='--', color='0.5', y=np.mean(chain.P[1:-1]), linewidth=2.0)
					axes[a].set_ylabel(r'Periodicity', fontsize=AXESFONTSIZE)
					axes[a].set_ylim(3.3, 3.9)
					axes[a].yaxis.set_ticks(np.arange(3.3, 3.9, 0.1))

				elif e=='Radius':
					# Plot radius
					axes[a].plot(chain.radius[1:-1], c=color, linewidth=1.0)
					axes[a].set_ylabel(r'Radius', fontsize=AXESFONTSIZE)
					axes[a].set_ylim(6, 9)
					axes[a].yaxis.set_ticks(np.arange(5, 9, 0.5))

				elif e=='CrickDev':
					# Plot axial rotation
					axes[a].plot(chain.crdev[1:-1], c=color, linewidth=1.0)
					axes[a].set_ylabel(r'Helix axial rotation', fontsize=AXESFONTSIZE)
					axes[a].set_ylim(-26, 26)
					axes[a].yaxis.set_ticks(np.arange(-26, 26, 10))

				elif e=='Shift':
					# Plot axial shift
					axes[a].plot(chain.shift[1:-1], c=color, linewidth=1.0)
					axes[a].set_ylabel(r'Axial shift', fontsize=AXESFONTSIZE)
					axes[a].set_ylim(-5, 5)
					axes[a].yaxis.set_ticks(np.arange(-5, 5, 1))

				elif e=='w0':
					axes[a].plot([i.w0 for i in chain.res[1:-1]], c=color, linewidth=1.0)
					axes[a].set_ylabel(r'w0', fontsize=AXESFONTSIZE)
					axes[a].set_ylim(-5, 5)
					axes[a].yaxis.set_ticks(np.arange(-5, 5, 1))

				elif e=='w1':
					axes[a].plot([i.w1 for i in chain.res[1:-1]], c=color, linewidth=1.0)
					axes[a].set_ylabel(r'w1', fontsize=AXESFONTSIZE)
					axes[a].set_ylim(90, 110)
					axes[a].yaxis.set_ticks(np.arange(90, 120, 5))

				else:
					pass

				if a == (len(elements)-1):
					axes[a].set_xlabel(r'Position in the sequence', fontsize=AXESFONTSIZE)
			a+=1

		for ax in axes:
			ax.xaxis.set_ticks(np.arange(0,len(self.layers)-2,1))

			for tick in ax.yaxis.get_major_ticks():
				tick.label.set_fontsize(TICKFONTSIZE)

		if filename=="":
			plt.show()
		else:
			plt.savefig(filename, dpi=150)

	def pymol_plot_layer(self, filename, savepath, suffix, pymol_version, color_selection=False, helix_order=False, helices_axis=None):
		"""Save layers to pymol session.

		Arguments:
			filename: path to pdb file
			savepath: path to where pse file will be saved
			suffix: suffix appended to the name of saved file (pdb_id)
			pymol_version: version of pymol used to generate session (differences in syntax between 1.x and 2.x)
			color_selection: should bundles be highlighted in pymol session

		Returns:
			None, session is saved to file
		"""

		# plot only interaction between helices that were used to determine layers
		if helix_order:
			h_order = self.helix_order
		else:
			h_order = None

		# plot layers by axis points
		layer_points = [ layer.get_layer_axis() for layer in self.layers if (len(layer.get_layer_axis()) == len(layer.res)) ]

		if color_selection:
			save_layers_to_pymol(filename, layer_points, savepath, suffix, pymol_version, color_selection=self.pymol_selection,
								 helix_order=h_order, ppo=self.ppo, bundle_axis=self.axis, helices_axis=self.helices_axis)
		else:
			save_layers_to_pymol(filename, layer_points, savepath, suffix, pymol_version, color_selection=False,
								 helix_order=h_order, ppo=self.ppo, bundle_axis=self.axis, helices_axis=self.helices_axis)



class residueClass():
	"""Class describing single residue"""

	def __init__(self, res, chain):
		self.res = res
		self.Ca = self.res['CA'].get_vector()
		self.chain_name = chain
		self.C = None
		self.O = None



### wrappers for external software ###

import subprocess
import tempfile

def run_dssp(pdbpath, dssp_binpath):
	"""Run DSSP software.
	Only 2.2.x versions are compatibile with Socket.

	Arguments:
	pdbpath      - path to PDB file
	dssp_binpath - path to dssp bin file
	
	Returns:
	Filepath to created .dssp file
	"""

	f = tempfile.NamedTemporaryFile(dir='/tmp/', mode='wt', delete=False)
	f.close()

	outfile_name = f.name

	cmd = f"{dssp_binpath} -i {pdbpath} -o {outfile_name}"

	try:
		subprocess.run(cmd, shell=True, check=True)
		return outfile_name
	except subprocess.CalledProcessError:
		print('SamCC-Turbo requires dssp 2.2.1 to run. Specify path in bin_paths')

def run_socket(pdbpath, dssppath, socket_binpath, cut_off=7.4):
	"""Run Socket using pdb and calculated dssp file

	Arguments:
	pdbpath        - path to PDB file
	dssppath       - path to .dssp file
	socket_binpath - path to Socket bin file
	cut_off        - Socket cutoff parameter (max distance where knobes-into-holes are detected)

	Returns:
	Filepath to created Socket output file

	"""

	f = tempfile.NamedTemporaryFile(dir='/tmp/', mode='wt', delete=False)
	f.close()

	outfile_name = f.name

	cmd = "%s -c %s -l -f %s -s %s > %s" % (socket_binpath, cut_off, pdbpath, dssppath, outfile_name)

	try:
		subprocess.run(cmd, shell=True, check=True)
		return outfile_name
	except subprocess.CalledProcessError:
		print('SamCC-Turbo requires Socket 3.03 to run. Specify path in bin_paths')



### classes for storing bundle as a collection of helix axis and points on helix axis ###

import itertools
import heapq
import numpy as np
from scipy.spatial import distance
from .chainClass import chainClass

class NotEqualAxisLen(Exception):
	pass

class helixAxisBundleClass():
	### class for storing all helices axis

	def __init__(self, chains, mode='from_chain', neighbour_interactions=None, ppo=None, average_dist_angle=None,
				 average_layers_dist=None, average_dist_to_plane=None, ranks=0):
		if mode == 'from_chain':
			self.helix_axis_all = [ helixAxisClass(helix_axis, helix_id=helix_id) for helix_id, helix_axis in enumerate(chains) ]
			self.neighbour_interactions = neighbour_interactions
			self.ppo = ppo
			self.average_dist_angle = average_dist_angle
			self.average_layers_dist = average_layers_dist
			self.average_dist_to_plane = average_dist_to_plane
			self.ranks = ranks
		elif mode == 'from_slice':
			self.helix_axis_all = chains
			self.neighbour_interactions = neighbour_interactions
			self.ppo = ppo
			self.average_dist_angle = average_dist_angle
			self.average_layers_dist = average_layers_dist
			self.average_dist_to_plane = average_dist_to_plane
			self.ranks = ranks

	def __getitem__(self, index):
		"""return new helixAxisBundleClass object but truncated
		when slicing only points with distance_flag = True are considered
		"""

		if isinstance(index, slice):
			helix_axis_sliced = []
			for helix_axis in self.helix_axis_all:
				# select only point with distance_flag = True and slice according to index
				helix_axis_sliced.append(helixAxisClass([ p for p in helix_axis.axis_points if p.distance_flag ][index]))

			return helixAxisBundleClass(helix_axis_sliced, mode='from_slice', neighbour_interactions=self.neighbour_interactions,
										ppo=self.ppo, average_dist_angle=self.average_dist_angle, average_layers_dist=self.average_layers_dist,
										average_dist_to_plane=self.average_dist_to_plane, ranks=self.ranks)
		else:
			return self.helix_axis_all[index]

	def iterchain(self):
		return self.helix_axis_all

	def iterlayer(self):
		"""Iterate over layers of the bundle, each item is one layer (list of points)
		Works only if all axis of equal length, else raise error
		"""

		# check if all helix axis of equal len, else raise error
		if not all(i.length() == self.helix_axis_all[0].length() for i in self.helix_axis_all):
			raise NotEqualAxisLen('Not all helix axis in the bundle are of equal length')

		# construct list of layers
		layers_all = []
		for l in range(self.helix_axis_all[0].length()):
			layer = []
			for ha in range(len(self.helix_axis_all)):
				layer.append(self.helix_axis_all[ha][l])
			layers_all.append(layer)

		return layers_all

	def get_middle_points(self, slice_size):
		"""for each helix take n/n-1 points from middle of a given helix; n when n is even and n-1 when n is uneven"""

		helix_axis_sliced = []
		for helix_axis in self.helix_axis_all:
			# select only point with distance_flag = True and slice according to index
			points_flag  = [ p for p in helix_axis.axis_points if p.distance_flag ]
			middle_point = int(len(points_flag) / 2)
			helix_axis_sliced.append(helixAxisClass(points_flag[middle_point - int(slice_size/2):middle_point + int(slice_size/2)]))

		return helixAxisBundleClass(helix_axis_sliced, mode='from_slice', neighbour_interactions=self.neighbour_interactions,
									ppo=self.ppo, average_dist_angle=self.average_dist_angle, average_layers_dist=self.average_layers_dist,
									average_dist_to_plane=self.average_dist_to_plane, ranks=self.ranks)

	def convert_to_coords_list(self):
		"""Return list of lists[helix axis] of lists[point coords]"""

		helix_axis_convert = []
		for helix_axis in self.helix_axis_all:
			helix_axis_convert.append([ [ c for c in p.coords ] for p in helix_axis.axis_points if (p.distance_flag and p.coords != []) ])

		return helix_axis_convert

	def detect_helix_order(self, res_num_layer_detection_asserted):

		def filter_out_diagonals_angles(first_points_on_helix):

			def calc_plane_bis(x, y, z):
				a = np.column_stack((x, y, z))
				return np.linalg.lstsq(a, np.ones_like(x), rcond=None)[0]

			def project_points(x, y, z, a, b, c):
				"""Projects the points with coordinates x, y, z onto the plane
				defined by a*x + b*y + c*z = 1
				"""
				vector_norm = a*a + b*b + c*c
				normal_vector = np.array([a, b, c]) / np.sqrt(vector_norm)
				point_in_plane = np.array([a, b, c]) / vector_norm

				points = np.column_stack((x, y, z))
				points_from_point_in_plane = points - point_in_plane
				proj_onto_normal_vector = np.dot(points_from_point_in_plane,
												 normal_vector)
				proj_onto_plane = (points_from_point_in_plane -
								   proj_onto_normal_vector[:, None]*normal_vector)

				return point_in_plane + proj_onto_plane

			def unit_vector(vector):
				""" Returns the unit vector of the vector.  """
				return vector / np.linalg.norm(vector)

			def angle_between(v1, v2):
				""" Returns the angle in radians between vectors 'v1' and 'v2'::

						>>> angle_between((1, 0, 0), (0, 1, 0))
						1.5707963267948966
						>>> angle_between((1, 0, 0), (1, 0, 0))
						0.0
						>>> angle_between((1, 0, 0), (-1, 0, 0))
						3.141592653589793
				"""
				v1_u = unit_vector(v1)
				v2_u = unit_vector(v2)
				return np.rad2deg(np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0)))

			def determine_helix_order(points_projected):

				helix_order          = [0] # set first helix (from socket data) as first helix
				current_helix_vector = points_projected[0,:] # get point representing first helix

				while(len(helix_order) < len(points_projected)):
					angles = {}
					for other_vector in enumerate(points_projected):
						if other_vector[0] not in helix_order:
							angles[other_vector[0]] = angle_between(current_helix_vector, other_vector[1])
					min_angle = min(angles, key=angles.get)
					helix_order.append(min_angle)
					current_helix_vector = points_projected[min_angle,:]

				return helix_order

			def make_list_of_neighbour_interactions(helix_order):

				neighbour_interactions = []

				for i in range(len(helix_order) - 1):
					neighbour_interactions.append((helix_order[i], helix_order[i+1]))
				neighbour_interactions.append((helix_order[0], helix_order[-1]))

				# correct order of helices (smaller id must be always first - this is format used later)
				neighbour_interactions = [ (i[0], i[1]) if i[0] < i[1] else (i[1], i[0]) for i in neighbour_interactions ]

				return neighbour_interactions

			# process points data
			x = [ point[0] for point in first_points_on_helix ]
			y = [ point[1] for point in first_points_on_helix ]
			z = [ point[2] for point in first_points_on_helix ]

			points_projected = project_points(x, y, z, *calc_plane_bis(x, y, z))

			ppo = project_points(x, y, z, *calc_plane_bis(x, y, z))

			# set center of coordinate system at centroid point
			points_projected = points_projected[:-1,:] - points_projected[-1,:]

			points_distances    = distance.cdist(points_projected, points_projected)
			max_points_distance = points_distances[points_distances>0].max() - points_distances[points_distances>0].min()

			helix_order            = determine_helix_order(points_projected)
			neighbour_interactions = make_list_of_neighbour_interactions(helix_order)

			return neighbour_interactions, max_points_distance, ppo

		def calculate_and_add_centroid(first_points_on_helix):

			average_x = np.mean([point[0] for point in first_points_on_helix])
			average_y = np.mean([point[1] for point in first_points_on_helix])
			average_z = np.mean([point[2] for point in first_points_on_helix])

			centroid = [average_x, average_y, average_z]
			first_points_on_helix.append(centroid)

			return first_points_on_helix

		def find_closest_first_points_on_helix_to_centroid(centroid, helices_pts_all):
			"""find a set of points so each point will be on different helix and distance
			from this point to centroid will be smallest for entire helix
			"""

			first_points_on_helix = []
			for helix in enumerate(helices_pts_all):
				dist = distance.cdist([centroid], helix[1])
				first_points_on_helix.append(helices_pts_all[helix[0]][dist[0].argmin()])

			first_points_on_helix.append(centroid)
			return first_points_on_helix

		middle_pts = self.get_middle_points(res_num_layer_detection_asserted)

		helices_pts_set = (middle_pts.convert_to_coords_list(),)
		helices_pts_all = self[:].convert_to_coords_list()

		largest_points_distance     = []
		neighbour_interactions_list = []
		ppo = [] # points used to determine helix order

		for helices_pts in helices_pts_set:
			pts_distances = {} # dict of distances between points on helices axes, key=(h_id1,h_id2)

			for h1 in enumerate(helices_pts):
				for h2 in enumerate(helices_pts):
					if (h1[0] != h2[0] and not (h2[0],h1[0]) in pts_distances.keys()):
						dist = distance.cdist(h1[1], h2[1]) # Compute distance between each pair of the two collections of inputs.
						pts_distances[(h1[0],h2[0])] = dist
			dist_matrices_number = len(pts_distances.keys()) # number of all possible distances (including diagonals)

			# mapping of distance id to record in pts_distances
			distance_to_pos = { pos:list(pts_distances.keys())[pos] for pos in range(dist_matrices_number)}
			# select only relations between neighbouring helices - new angle-based method

			first_points_on_helix = [ helix[int(len(helix)/2)] for helix in helices_pts ]
			first_points_on_helix = calculate_and_add_centroid(first_points_on_helix)

			# find first_points_on_helix from centroid
			first_points_on_helix = find_closest_first_points_on_helix_to_centroid(first_points_on_helix[-1], helices_pts_all)

			neighbour_interactions, max_points_distance, ppo = filter_out_diagonals_angles(first_points_on_helix)
			ppo = first_points_on_helix
			largest_points_distance.append(max_points_distance)
			neighbour_interactions_list.append(neighbour_interactions)

		best_order = largest_points_distance.index(min(largest_points_distance))

		# add to self:
		self.neighbour_interactions = neighbour_interactions_list[best_order]
		self.ppo = ppo

	def get_all_bundle_boundry_layers(self):
		"""Only get list of searchLayers - do not evaluate best ones at this point"""

		return [ searchLayer(l, self.neighbour_interactions) for l in itertools.product(*self.helix_axis_all, repeat=1) ]

	def find_bundle_boundry_layer(self, distance_threshold, search_layer_setting_num):

		# use searchLayer class to create layer settings and calculate distance between their points
		layers_all = [ searchLayer(l, self.neighbour_interactions) for l in itertools.product(*self.helix_axis_all, repeat=1) ]

		# get set of layers with n smallest distances between their points
		min_distance_set = heapq.nsmallest(search_layer_setting_num, layers_all)

		# check if smallest distance between layers to large for reasonable layer
		# return only layers that meet reasonable distance criteria
		return [ layer for layer in min_distance_set if layer.total_distance <= distance_threshold ]

	def find_all_layers_from_layer(self, boundry_layers):

		layer_sets = []

		for layer in boundry_layers:
			helix_data = []
			for helix_axis, layer_point in zip(self.helix_axis_all, layer):
				# search helix axis for boundry point and at the same time calculate how many points upstream and downstream from it
				pts_before = 0
				pts_after  = 0
				pt_found   = False
				for axis_point in helix_axis:

					if axis_point.distance_flag == False:
						continue
					if axis_point.point_id == layer_point.point_id:
						pt_found = True
					# point before investigated one
					elif pt_found == False:
						pts_before += 1
					elif pt_found == True:
						pts_after += 1
				helix_data.append([pts_before, pts_after])
			max_upstream   = min([h[0] for h in helix_data])
			max_downstream = min([h[1] for h in helix_data])

			# extract axis points and construct helixAxisBundleClass:
			helix_axis_sliced = []
			for helix_axis, layer_point in zip(self.helix_axis_all, layer):
				idx_start = layer_point.point_id - max_upstream
				idx_stop  = layer_point.point_id + max_downstream + 1
				helix_axis_sliced.append(helix_axis[idx_start:idx_stop])
			layer_sets.append(helixAxisBundleClass(helix_axis_sliced, mode='from_slice', neighbour_interactions=self.neighbour_interactions,
												   ppo=self.ppo, average_dist_angle=self.average_dist_angle, average_layers_dist=self.average_layers_dist,
												   average_dist_to_plane=self.average_dist_to_plane, ranks=self.ranks))

		return layer_sets

	def verify_points_distances(self, max_distance=20):
		"""flag axis points that does not meet distance criteria
		code verifying if all helix axis points have any axis points
		from other helix within reasonable distance (20A) [dev-doc]
		this code compares all vs all axis points from all helices
		"""

		for h1_idx, helix_axis1 in enumerate(self.helix_axis_all):
			for point1 in helix_axis1.axis_points:
				# if edge (first or last on helix) flag as False
				if point1.coords == []:
					point1.distance_flag = False
					continue
				point_stat = []
				for h2_idx, helix_axis2 in enumerate(self.helix_axis_all):
					if h1_idx != h2_idx:
						for point2 in helix_axis2.axis_points:
							if (any(point1.CA_coords) and any(point2.CA_coords)):
								dst = distance.euclidean(point1.CA_coords, point2.CA_coords)
								if dst < max_distance:
									# found residue within cutoff distance on this helix, break and search next helix
									point_stat.append(True)
									break
						else:
							# there is no residue within cutoff distance on this helix, search next one
							point_stat.append(False)

				if all(point_stat):
					# there is residue within cutoff distance on every other helix - set distance_flag to True
					point1.distance_flag = True
				else:
					# set flag to False: at least one helix is too far away from this point
					point1.distance_flag = False

	def show_points(self):
		print('Bundle of', len(self.helix_axis_all), 'axis')
		print('Helix order:', self.neighbour_interactions)
		print('Average distance/angle of layers:', self.average_dist_angle)
		print('Average total distance of layers:', self.average_layers_dist)
		print('Average distance to plane of points in layers:', self.average_dist_to_plane)
		print('Layer set aggregated rank:', self.ranks)
		for hid, helix in enumerate(self.helix_axis_all):
			print('Axis', hid)
			for r in helix.axis_points:
				print('Point', r.point_id  ,'from:', r.res, ', dist_flag:', r.distance_flag, ', chain:', r.chain_name, '(' + str(r.helix_id) + ')')
			print('---'*5)

class helixAxisClass():
	"""class for storing helix axis - composed of helix points"""

	def __init__(self, helix_axis, helix_id=None):
		if isinstance(helix_axis, chainClass):
			self.axis_points = [ axisPoint(pid,data[0],data[1],helix_id) for pid, data in enumerate(zip(helix_axis.axis, helix_axis.res)) ]
		elif isinstance(helix_axis, list):
			self.axis_points = helix_axis

	def __repr__(self):
		return 'Chain of len ' + str(len(self.axis_points))

	def __getitem__(self, index):
		if isinstance(index, slice):
			return helixAxisClass(self.axis_points[index])
		else:
			return self.axis_points[index]

	def length(self, dst_flag=False):
		if dst_flag:
			return len([ p for p in self.axis_points if p.distance_flag ])
		else:
			return len(self.axis_points)

class axisPoint():
	"""class for storing helix points"""

	def __init__(self, pid, axis_point, residue, helix_id):
		self.point_id = pid
		if axis_point:
			self.coords = axis_point.get_array()
		else:
			self.coords = []
		self.CA_coords = residue.Ca.get_array()
		self.distance_flag = None
		self.res = residue.res
		self.chain_name = residue.chain_name
		self.helix_id   = helix_id

	def __repr__(self):
		if self.coords != []:
			return 'Axis point {0} at {1:.2f},{2:.2f},{3:.2f}'.format(self.point_id, *[float(c) for c in self.coords])
		else:
			return 'Axis point {0} at [no coords]'.format(self.point_id)

class searchLayer():
	"""class for storing layer data in search for min distance layers"""

	def __init__(self, layer_points, neighbour_interactions):
		self.axis_points    = [ l for l in layer_points ]
		self.total_distance = self.calculate_total_distance(neighbour_interactions)

	def __repr__(self):
		return 'Layer# Tot dist: {0:.2f}  Point ids: {1} Res ids: {2}'.format(self.total_distance, str([ p.point_id for p in self.axis_points ]), str([ p.res.get_id()[1] for p in self.axis_points ]))

	def __getitem__(self, index):
		return self.axis_points[index]

	def __lt__(self, other):
		return self.total_distance < other.total_distance

	def calculate_total_distance(self, neighbour_interactions):
		"""Calculate total distance between neighbouring points in layer
		neighbourhood is determined according to neighbour_interactions list
		"""

		total_distance = 0

		for axis_point1 in self.axis_points:
			for axis_point2 in self.axis_points:
				# if (axis_point1.helix_id != axis_point2.helix_id):
				if (axis_point1.helix_id, axis_point2.helix_id) in neighbour_interactions:
					total_distance += distance.euclidean(axis_point1.coords, axis_point2.coords)

		return total_distance



def save_layers_to_pymol(pdbpath, layer_points, savepath, suffix, pymol_version=2.0, helices_axis=None,
						 color_selection=False, helix_order=None, ppo=None, bundle_axis=False):
	''' save each layer as a set of distances between point determining the layer '''

	import __main__
	__main__.pymol_argv = [ 'pymol', '-qc'] # Quiet and no GUI
	import sys, time, os, itertools
	try:
		import pymol
	except ModuleNotFoundError:
		import sys
		print('Module pymol not found')
		print('Pymol required to save pse files. Add pymol to your pythonpath or run SamCC-Turbo with save_pse=False flag')
		sys.exit()

	pymol.finish_launching()
	pymol.cmd.reinitialize()

	pdbname = pdbpath.split('/')[-1].split('.')[0]

	pymol.cmd.load(pdbpath, pdbname)

	# this draws points that are used to find helix order
	for p in enumerate(ppo):
		pymol.cmd.pseudoatom('ppo_' + str(p[0]), pos=list(p[1]))

	### axis drawing code ###
	# helices axis
	if helices_axis:
		for helix in enumerate(helices_axis):

			h_start  = list(helix[1][1])
			h_end    = list(helix[1][-2])
			h_name_s = 'axp' + str(helix[0]) + '_start'
			h_name_e = 'axp' + str(helix[0]) + '_end'

			pymol.cmd.pseudoatom(h_name_s, pos=h_start)
			pymol.cmd.pseudoatom(h_name_e, pos=h_end)

			### this version works for pymol 1.7
			# pymol.cmd.distance('dist' + str(helix[0]), '/' + h_name_s + '/////1', '/' + h_name_e + '/////1')
			### this version works for pymol 2.0
			pymol.cmd.distance('axis' + str(helix[0]), '/' + h_name_s + '///1', '/' + h_name_e + '///1')
	# bundle axis
	b_axis_start = list(bundle_axis[1])
	b_axis_end   = list(bundle_axis[-2])
	b_name_start = 'bundle_axis_start'
	b_name_end   = 'bundle_axis_end'

	pymol.cmd.pseudoatom(b_name_start, pos=b_axis_start)
	pymol.cmd.pseudoatom(b_name_end, pos=b_axis_end)

	pymol.cmd.distance('bundle_axis', '/' + b_name_start + '///1', '/' + b_name_end + '///1')
	### === ###

	### only show selection and color it code ###
	if color_selection:
		pymol.cmd.hide('cartoon', pdbname)
		pymol.cmd.show('cartoon', ' or '.join(color_selection))
		pymol.cmd.color('red', ' or '.join(color_selection))
	### === ###

	### layer creation code
	for layer in enumerate(layer_points):
		for point in enumerate(layer[1]):
			pymol.cmd.pseudoatom('l_' + str(layer[0]) + 'point' + str(point[0]), pos=point[1])

		for dist in itertools.combinations(range(len(layer[1])), 2):
			# draw only interactions between neighbouring helices (if helix_order not None)
			if (helix_order and (dist[0], dist[1]) not in helix_order):
				continue
			if pymol_version == 2.0:
				pymol.cmd.distance('l_' + str(layer[0]) + 'dist_' + str(dist[0]) + '_' + str(dist[1]), '/l_' + str(layer[0]) + 'point' + str(dist[0]) + '///1', '/l_' + str(layer[0]) + 'point' + str(dist[1]) + '///1')
			else:
				pymol.cmd.distance('l_' + str(layer[0]) + 'dist_' + str(dist[0]) + '_' + str(dist[1]), '/l_' + str(layer[0]) + 'point' + str(dist[0]) + '/////1', '/l_' + str(layer[0]) + 'point' + str(dist[1]) + '/////1')
	### end layer creation

	pymol.cmd.set('dash_gap', 0)
	pymol.cmd.set('dash_radius', 0.2)
	pymol.cmd.hide('labels')
	pymol.cmd.hide('lines')
	#pymol.cmd.show('cartoon')
	pymol.cmd.remove('resn HOH')

	print('Saving pymol session with layers...')
	pymol.cmd.save(savepath + '/' + pdbname + '_' + suffix + '.pse', format='pse')

	print(pdbname + ' saved.')



### Main SamCC-Turbo file ###
DEBUG=False

import pickle
import sys, os
import pathlib
from .wrappers import run_dssp
from .wrappers import run_socket
from .socket_parser import parse_socket_output
from .socketClass import socket_class
from .bundleClass import bundleClass

class SamCC():
	"""main SamCC-Turbo class"""

	def __init__(self, bin_paths={'dssp':'dssp', 'socket':'socket'}):

		self.bin_paths = bin_paths

	def run_samcc_turbo(self, pdbpath, outpath, mode='auto-detect', defdata=None,
						plot=True, save_df=True, save_pse=True,
						layer_detect_n=5, max_dist='auto', search_set_n=9):
		"""Main function for running samcc-turbo.

		Arguments (general):
		pdbpath   -- path to pdb file (.pdb)
		outpath   -- path where the output files will be stored
		mode      -- mode of running SamCC (default 'auto-detect')
		           - auto-detect: automatic detection of layers in the bundle (requires
		                          installed dssp and Socket)
				   - defdata: use layer definition from list
		plot      -- plot results and also save plot to file (default True)
		save_df   -- save result DataFrame to pickle (default True)
		save_pse  -- save pymol session with drawn layers (default True)
		bin_paths -- dictionary of paths to binaries of dssp and socket
		             (default {'dssp': 'dssp', 'socket':'socket'})

		Arguments (auto-detect mode specific):
		layer_detect_n -- number of residues from each helix that will be combined
						  into layers that will be considered in layer detection;
						  (default 5) [this is "r" parameter in the paper]
		max_dist       -- distance threshold [A] that will be used to indicate unusual
						  layer setting; if total distance between residues in layer
						  if greater that this number the layer will be discarded, if
						  no layer will pass this filter then layer detection will
						  be repeated with bigger layer_detect_n; if set to auto
						  then threshold will be assigned basing on oligomerization
						  (default auto)
		search_set_n   -- number of starting layer settings that will be selected
						  basing on minimal distance between residues; they will be compared
						  in terms of angle between all layers in structure (default 9)
		Arguments (defdata mode specific):
		defdata   -- list of parameters defining layer setting (default None)
		"""

		assert os.path.isdir(outpath), f'"{outpath}" should be a valid directory'
		if outpath[-1]!='/':outpath+='/'

		pdbid = pathlib.Path(pdbpath).stem

		if mode == 'auto-detect':
			# run dssp and socket ('auto-detect' mode)
			dssppath       = run_dssp(pdbpath, self.bin_paths['dssp'])
			socketpath     = run_socket(pdbpath, dssppath, self.bin_paths['socket'])
			socket_data    = parse_socket_output(socketpath)
			s 			   = socket_class(socket_data, pdbpath)

			os.remove(dssppath)
			os.remove(socketpath)

			bundles = s.get_bundles(res_num_layer_detection=layer_detect_n,
									distance_threshold=max_dist,
									search_layer_setting_num=search_set_n)

			for bid, bundle in enumerate(bundles):
				bundle.calc_bundleaxis()
				bundle.get_helicesaxis()
				bundle.calc_periodicity()
				bundle.calc_radius()
				bundle.calc_crick()
				bundle.assign_positions()
				bundle.calc_crickdev(3.5, 7, optimal_ph1=19.5)
				bundle.calc_axialshift()

				out_file = f'{outpath}{pdbid}_{bid}'

				if plot: # make plot and save it to file
					bundle.plot(out_file + '.png', elements=['Periodicity', 'Radius', 'CrickDev', 'Shift'])

				if save_df: # dump pickle with dataframe of measured values
					pickle.dump(bundle.gendf(), open(out_file + '.p', 'wb'))

				if save_pse:
					bundle.pymol_plot_layer(filename=pdbpath, savepath=outpath, suffix=str(bid),
											pymol_version=2.0, color_selection=True, helix_order=bundle.helix_order, helices_axis=bundle.helices_axis)

		elif mode == 'defdata':

			bundle = bundleClass()
			bundle.from_defdata(pdbpath, *defdata)

			bundle.calc_bundleaxis()
			bundle.get_helicesaxis()
			bundle.calc_periodicity()
			bundle.calc_radius()
			bundle.calc_crick()
			bundle.assign_positions()
			bundle.calc_crickdev(3.5, 7, optimal_ph1=19.5)
			bundle.calc_axialshift()

			out_file = f'{outpath}{pdbid}_0'

			if plot: # make plot and save it to file
				bundle.plot(out_file + '.png', elements=['Periodicity', 'Radius', 'CrickDev', 'Shift'])

			if save_df: # dump pickle with dataframe of measured values
				pickle.dump(bundle.gendf(), open(out_file + '.p', 'wb'))

			if save_pse:
				bundle.pymol_plot_layer(filename=pdbpath, savepath=outpath, suffix='0',
										pymol_version=2.0, color_selection=True, helix_order=bundle.helix_order, helices_axis=bundle.helices_axis)

		else:
			print('Unknown mode. Available modes: auto-detect and defdata')



### socketClass definition ###

import numpy as np
from .helper_functions import detect_helices_orientation
from .searchLayerClass import helixAxisBundleClass
from .residueClass import residueClass
from .chainClass import chainClass
from .bundleClass import bundleClass
from .layerClass import layerClass
from Bio.PDB import PDBParser
from .layer_detection import select_minimal_distance_layer_set
from .layer_detection import select_minimal_angle_layer_set2
from .layer_detection import select_minimal_angle_layer_set
from .layer_detection import find_bundle_boundry_layer_from_all
from .layer_detection import create_pymol_selection_from_socket_results
from .layer_detection import select_minimal_total_distance_layer_set
from .layer_detection import select_minimal_dist_to_plane_layer_set
from .layer_detection import get_layers_set_with_best_ranks
from .layer_detection import find_closest_CA_to_point

### Exceptions

class NoResidue(Exception):
	pass

class TooShort(Exception):
	pass

class socket_class():
	"""Class responsible for handling Socket-derived data"""

	def __init__(self, socket_data, pdb_filename):

		self.socket_data = socket_data
		self.pdb_filename = pdb_filename

	def get_bundles(self, res_num_layer_detection=5, distance_threshold='auto', search_layer_setting_num=9):
		"""Given a SOCKET raw data calculates a proper bundle(s) in which all
		helices are of the same length and residues are arrranged in layers.

		Arguments (same description as in main SamCC-Turbo function):
		res_num_layer_detection   -- number of residues from each helix that will be combined
						  		     into layers that will be considered in layer detection;
						             (default 5) [this is "r" parameter in the paper]
		distance_threshold        -- distance threshold [A] that will be used to indicate unusual
							  		 layer setting; if total distance between residues in layer
									 if greater that this number the layer will be discarded, if
									 no layer will pass this filter then layer detection will
									 be repeated with bigger layer_detect_n; if set to auto
									 then threshold will be assigned basing on oligomerization
									 (default auto)
		search_layer_setting_num  -- number of starting layer settings that will be selected
									 basing on minimal distance between residues; they will be compared
									 in terms of angle between all layers in structure (default 9)

		Returns:
			bundles: list of bundleClass instances (one structure can contain multiple bundles)
		"""

		def parse_socket():
			"""Parse raw SOCKET data and generate a list of helices

			Arguments:
			socket_data -- raw SOCKET data

			Returns:
			bundles: list of bundles (i.e list of helices: from, to, chain, anti-parallel)
			"""

			bundles = []

			for cc_id in self.socket_data:
				cc = self.socket_data[cc_id]
				# detect helices orientation
				o = detect_helices_orientation(cc['indices'], cc['relations'])

				bundleDesc = []
				for helix_id in cc['indices']:
					helix = cc['indices'][helix_id]
					bundleDesc.append([helix['start'], helix['end'], helix['chain'], o[helix_id]==-1])

				# if all helices in a bundle are AP then we should
				# switch them all to P (parallel)

				if all([i[3] for i in bundleDesc]):
					new_bundleDesc=[]
					for i in bundleDesc:
						new_bundleDesc.append(i[:-1] + [False])
					bundleDesc = new_bundleDesc

				bundles.append(bundleDesc)

			return bundles

		def parse_pdb(input_helices, pdb_filename):
			"""
			Parses PDB file according to SOCKET definitions

			Arguments:
				input_helices: list of helices (from, to, chain, ap)

			Returns:
				chains: list of chainClass instances
						(note that these chains may have different lengths)
			"""

			parser=PDBParser(QUIET=True)
			structure=parser.get_structure('temp', pdb_filename)

			chains = []

			for helix in input_helices:
				start, end, chain, ap = helix

				# solve the problem with het atoms
				preres = {}
				try:
					for i in structure[0][chain]:
						j = i.get_full_id()
						preres[j[3][1]] = i
				except KeyError:
					raise NoResidue()

				try:
					residues = [residueClass(preres[r], chain) for r in range(start, end+1)]
				except KeyError:
					raise NoResidue

				if ap == True:
					residues = residues[::-1]

				#print(len(residues), ap, chain)
				chain = chainClass(residues, ap, chain)
				chain.new_starts=[]
				chains.append(chain)

			assert len(chains)>1
			return chains

		bundles = []

		for input_helices in parse_socket():

			try:
				chains = parse_pdb(input_helices, self.pdb_filename)
			except NoResidue:
				print('No residue error')
				continue

			# assign distance_threshold according to oligomerization state
			default_distance_threshold = {2:30, 3:40, 4:50, 5:60, 6:70, 7:80, 8:90, 9:100}
			if distance_threshold == 'auto':
				distance_threshold_set = default_distance_threshold[len(input_helices)]
			else:
				distance_threshold_set = distance_threshold

			# calculate axis of the bundle
			for c in chains:
				c.calc_axis(smooth=False)

			# initialize axis bundle from chain axis points
			helices_axis_all = helixAxisBundleClass(chains)

			# flag axis points that are too far away from other points
			helices_axis_all.verify_points_distances(max_distance=20)

			# initialize boundry_layers list and search scope
			boundry_layers                   = []
			res_num_layer_detection_asserted = res_num_layer_detection

			# condition: any layer (from all segments) has distance belowe threshold
			while(not boundry_layers):

				# determine helix order
				helices_axis_all.detect_helix_order(res_num_layer_detection_asserted)

				# subset axis points from start, middle and end of bundle (according to distance flag)
				helices_pts_first  = helices_axis_all[:res_num_layer_detection_asserted]
				helices_pts_last   = helices_axis_all[-res_num_layer_detection_asserted:]
				helices_pts_middle = helices_axis_all.get_middle_points(res_num_layer_detection_asserted)

				for helices_pts in [helices_pts_first, helices_pts_middle, helices_pts_last]:
					boundry_layers += helices_pts.get_all_bundle_boundry_layers()

				boundry_layers = find_bundle_boundry_layer_from_all(boundry_layers, distance_threshold=distance_threshold_set, search_layer_setting_num=search_layer_setting_num)

				# increase scope of searched residues
				# INCREMENT - to activate multiloop uncomment next line and comment out four following lines
				# res_num_layer_detection_asserted += 1
				if res_num_layer_detection_asserted == 15:
					break
				else:
					res_num_layer_detection_asserted = 15
				# INCREMENT END

			assert len(boundry_layers)>0

			# merge all boundry layers into one list - but only selected ones
			percentile_threshold = np.percentile([ bl.total_distance for bl in boundry_layers ], 50)
			boundry_layers       = [ bl for bl in boundry_layers if bl.total_distance < percentile_threshold ]

			layers_sets = helices_axis_all.find_all_layers_from_layer(boundry_layers)

			# dimer case
			if len(input_helices) == 2:
				best_layer_set = select_minimal_distance_layer_set(layers_sets)
			# all other cases
			else:
				best_layer_set = select_minimal_angle_layer_set(layers_sets, best_layer_nb=1)

			# convert best layer set to bundleClass()
			# -1/+1 to chain indexes to accomodate fact that chain edges are excluded from layer search
			# this is because it is not possible to calculate helix axis points for edge residues
			for c, b in zip(chains, best_layer_set.iterchain()):
				cut_start = b[0].point_id -1 # id of first residue in best layer chain
				cut_stop  = b[-1].point_id +1 # id of last residue in best layer chain
				c.res = c.res[cut_start:cut_stop+1]
				c.axis = c.axis[cut_start:cut_stop+1]

			bundle = bundleClass()

			# check whether all chains are of equal length
			temp = [len(i.res) for i in chains]
			assert all([i==temp[0] for i in temp])

			# final definition of helices
			bundle.chains = chains

			# generate layers
			bundle.layers = []
			for layer_residues in zip(*[c.res for c in bundle.chains]):
				bundle.layers.append(layerClass(layer_residues))

			# generate pymol-readable selection
			# this will cut range of shown helices to the residues that are within resonable distance from all other helices
			# it does not affect range of drawn layers in pymol
			input_helices_cut = []
			for ha, ih in zip(helices_axis_all[:].iterchain(), input_helices):
				if ih[3]:
					input_helices_cut.append([ha[-1].res.get_id()[1], ha[0].res.get_id()[1], ih[2], ih[3]])
				else:
					input_helices_cut.append([ha[0].res.get_id()[1], ha[-1].res.get_id()[1], ih[2], ih[3]])

			input_helices = input_helices_cut

			bundle.pymol_selection = create_pymol_selection_from_socket_results(input_helices)
			bundle.ppo = best_layer_set.ppo

			# save order of helices (helix indexes as in socket data)
			bundle.helix_order = best_layer_set.neighbour_interactions

			bundles.append(bundle)

		return bundles



import numpy as np
import itertools
from Bio.PDB import calc_dihedral
from Bio.PDB import vector_to_axis
from collections import deque
from .bundle import get_local_axis
from .helper_functions import diffangle
from .helper_functions import gen_expected_crick_angles
from .helper_functions import crick_to_pos
from .helper_functions import calc_crick_ang_dev
from .helper_functions import savitzky_golay

class chainClass():
	"""Class describing single chain (helix) of the structure"""

	def __init__(self, residues, ap, chain, heptad='x'):
		self.res = residues
		for r in self.res:
			r.parentChain = self
		self.ap = ap
		self.chain = chain
		self.heptad = heptad

	def __str__(self):
		return "%s [%s-%s]" % (self.chain, self.res[0].res.id[1], self.res[-1].res.id[1])

	def get_seq(self):
		"""Get chain sequence"""

		s = "".join([seq1(r.res.resname) for r in self.res])
		if self.ap: s=s[::-1]
		return s

	def get_CA(self):
		"""Returns a list of helices. Each helix is a list of CA atoms.
		Each CA atom is represented as 3-dimensional Biopython Vector object.
		"""

		return [r.Ca for r in self.res]

	def calc_axis(self, smooth=False):
		"""Calculate axis"""

		temp_axis = get_local_axis([i.Ca for i in self.res])

		if smooth:
			self.axis = [temp_axis[0]] + [np.sum(w[0]+Vector(w[1]._ar*2)+w[2])**(1.0/4) for w in window(temp_axis, n=3)] + [temp_axis[-1]]
		else:
			self.axis = temp_axis

		# None values at edges to assure equal length of parameters tables
		self.axis = [None] + self.axis + [None]

		for pos, ax in enumerate(self.axis):
			self.res[pos].O = ax

	def calc_radius(self):
		"""Calculates distance between helical axis points and bundle axis points"""

		for pos in range(1, len(self.res)-1):
			self.res[pos].radius = (self.res[pos].O - self.res[pos].C).norm()

		self.res[0].radius = None
		self.res[-1].radius = None

		radius = [r.radius for r in self.res[1:-1]]
		self.radius = [None]+radius+[None]

	def calc_periodicity(self, smooth=False, helix_p=3.63):
		"""Calculates bundle periodicity from the perspective of a helix

		Arguments:
			helix_p: None - use local helix periodicity for caluclations
				     value - use a given fixed value of helix periodicity (default=3.63; alpha-helix periodicity)
		"""

		for pos, r in enumerate(self.res[1:-1]):
			# real position in the array
			realpos=pos+1

			### Calculate helical phase yield per residue (w1)
			# help on dihedral angles:
			# https://math.stackexchange.com/questions/47059/how-do-i-calculate-a-dihedral-angle-given-cartesian-coordinates
			temp_w1 = []
			for other in [1, -1]:

				if (realpos+other>0) and (realpos+other<len(self.res)-1):
					temp_w1.append(np.degrees(calc_dihedral(r.Ca,\
															r.O,\
															self.res[realpos+other].O,\
															self.res[realpos+other].Ca)))

			self.res[realpos].w1 = np.mean(temp_w1)
			self.res[realpos].p = 360.0/self.res[realpos].w1

			### Calculate coiled coil phase yield per residue (w0)
			# https://github.com/numpy/numpy/issues/8529
			# Superhelical frequency (negative means left-handed superhelix)
			temp_w0 = []

			for other in [1, -1]:
				if (realpos+other>0) and (realpos+other<len(self.res)-1):
					with np.errstate(invalid="ignore"):
						temp_w0.append(np.degrees(calc_dihedral(r.O,\
																r.C,\
																self.res[realpos+other].C,\
																self.res[realpos+other].O)))

			self.res[realpos].w0 = np.mean(temp_w0)

			### Calculate bundle periodicity
			if helix_p==None: # use local helix periodicity (p) to calculate bundle periodicity (P)
				self.res[realpos].P = self.res[realpos].p / (1-(self.res[realpos].p*self.res[realpos].w0)/360)
			else: # use fixed p value
				self.res[realpos].P = helix_p / (1-(helix_p * self.res[realpos].w0)/360)

		self.res[0].P, self.res[-1].P = None, None
		self.p = [None]+[i.p for i in self.res[1:-1]]+[None]

		if smooth:
			self.P = [None]+list(savitzky_golay([r.P for r in self.res[1:-1]],5,1))+[None]
			# Update P in the residue objects
			for pos, i in enumerate(self.P):
				self.res[pos].P=i
		else:
			self.P = [None]+[r.P for r in self.res[1:-1]]+[None]

	def calc_crick(self):
		"""Calculates Crick angles for the chain."""

		def ClosestPointOnLine(a, b, p):
			ap = p-a
			ab = b-a
			result = a + np.dot(ap,ab)/np.dot(ab,ab) * ab
			return result

		def angle(v1, v2):
			return np.arccos(np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2)))

		for pos, r in enumerate(self.res[1:-1]):
			realpos=pos+1

			isFirstLayer = realpos==1
			isLastLayer = realpos==len(self.res)-2

			# parallel
			if not self.ap:
				if isFirstLayer:
					O_next = self.res[realpos+1].O.get_array()
					C_next = self.res[realpos+1].C.get_array()
				else:
					O_next = self.res[realpos-1].O.get_array()
					C_next = self.res[realpos-1].C.get_array()

			# anti-parallel
			else:
				if isLastLayer:
					O_next = self.res[realpos-1].O.get_array()
					C_next = self.res[realpos-1].C.get_array()
				else:
					O_next = self.res[realpos+1].O.get_array()
					C_next = self.res[realpos+1].C.get_array()

			O = r.O.get_array()
			Ca = r.Ca.get_array()
			C = r.C.get_array()

			# Project Ca onto the helix axis (O_new)
			O_new = ClosestPointOnLine(O, O_next, Ca)

			# Define a plane perpendicular to the helix axis
			# and find intersection between this plane and the bundle axis (C_new)
			n = O-O_new
			V0 = O_new
			w = C - O_new
			u = C_next - C
			N = -np.dot(n, w)
			D = np.dot(n, u)
			sI = N / D
			C_new = C+ sI * u

			# Define sign of the Crick angle
			mixed = np.dot(np.cross(O_next - O_new, Ca - O_new), C_new - O_new)

			if mixed<0:
				sign=1
			else:
				sign=-1

			if not self.ap:
				if not isFirstLayer: sign = sign * -1
			else:
				if not isLastLayer: sign = sign * -1

			r.crick = np.degrees(angle(Ca-O_new, C_new-O_new)) * sign

		self.res[0].crick=None
		self.res[-1].crick=None

		crick = [r.crick for r in self.res[1:-1]]
		self.crick = [None]+crick+[None]


	def old_calc_crick(self, fix_Crick=True):

		"""
		Currently not used!
		"""

		pass

		"""

		f_problem = []

		for pos, r in enumerate(self.res[1:-1]):

			realpos=pos+1

			OCa = r.Ca - r.O
			OC = r.C - r.O

			r.OCa = OCa
			r.OC = OC

			#Calculate the mixed product of vectors (O-Ca,O-C,O-Ox).
			#Just consider the vector between the current helical axis point
			#(O in your notation) and the next one (Ox).
			#If the mixed product is negative, reverse the sign of the Crick angle
			#(or vice versa, depending on definitions).

			if not self.ap:
				if realpos==1:
					OY = (self.res[realpos+1].O - r.O)
				else:
					OY = (self.res[realpos-1].O - r.O)
			else:
				if realpos==len(self.res)-2:
					OY = (self.res[realpos-1].O - r.O)
				else:
					OY = (self.res[realpos+1].O - r.O)

			mixed = (OY ** OCa) * OC

			if mixed<0:
				sign=1
			else:
				sign=-1

			if not self.ap:
				if realpos!=1: sign = sign * -1
			else:
				if realpos!=len(self.res)-2: sign = sign * -1

			r.crick = np.degrees(OCa.angle(OC)) * sign

			# defining the sign of the Crick angle close to the +/-180 deg. may be problematic
			# such cases are stored for further check-up
			if abs(mixed)<5 and abs(r.crick)>150:
				f_problem.append(realpos)

		self.res[0].crick=None
		self.res[-1].crick=None

		print(f_problem)

		if not fix_Crick: f_problem = []

		#iterate over potentially problematic Crick angles
		for fp in f_problem:

			expected = []

			# calculate w1 values in the neighbouring residue pairs
			if fp>2:
				expected.append(abs(diffangle(self.res[fp-1].crick, self.res[fp-2].crick)))
			if fp<len(self.res)-3:
				expected.append(abs(diffangle(self.res[fp+1].crick, self.res[fp+2].crick)))

			if len(expected)==0: continue

			expected = np.mean(expected)

			current = []
			alternative = []

			if fp>1:
				current.append(abs(diffangle(self.res[fp].crick, self.res[fp-1].crick)))
				alternative.append(abs(diffangle(-self.res[fp].crick, self.res[fp-1].crick)))
			if fp<len(self.res)-2:
				current.append(abs(diffangle(self.res[fp].crick, self.res[fp+1].crick)))
				alternative.append(abs(diffangle(-self.res[fp].crick, self.res[fp+1].crick)))

			if max(np.absolute(current - expected)) > max(np.absolute(alternative - expected)):
				print(fp, 'fixed!')
				self.res[fp].crick = -self.res[fp].crick

		crick = [r.crick for r in self.res[1:-1]]
		self.crick = [None]+crick+[None]

		"""

	def calc_pitch_angle(self):
		"""Calculate pitch angle"""

		for pos in range(len(self.res)-2):
			realpos=pos+1

			if realpos==1:
				other=1
			else:
				other=-1

			O = self.res[realpos].O - self.res[realpos+other].O
			C = self.res[realpos].C - self.res[realpos+other].C

			self.res[realpos].A = np.degrees(C.angle(O))
		self.res[0].A=None
		self.res[-1].A=None
		self.A = [None]+[r.A for r in self.res[1:-1]]+[None]

	def assign_positions(self):
		"""Assigns heptad positions using TWISTER algorithm."""

		hpos=['?', '?']
		pos = 2

		if self.ap:
			crick = self.crick[::-1]
		else:
			crick = self.crick

		while pos < len(self.res)-2:

			if crick[pos-1] < 0 and crick[pos] > 0 and abs(crick[pos-1]) > abs(crick[pos]):
				hpos.extend(['a', 'b', 'c'])
				pos+=3
			elif crick[pos] < 0 and crick[pos+1] > 0 and abs(crick[pos]) < abs(crick[pos+1]):
				hpos.extend(['d', 'e', 'f', 'g'])
				pos+=4
			else:
				hpos.append('?')
				pos+=1


		hpos = hpos[:len(self.res)-2]

		# add missing positions at N end
		fpos=0
		try:
			while hpos[fpos]=='?': fpos+=1
		except IndexError:
			pass
		else:
			hep = itertools.cycle('gfedcba')
			while not next(hep) == hpos[fpos]: pass
			fadd = [next(hep) for i in range(fpos)][::-1]
			hpos = fadd + hpos[fpos:]

		# add missing positions at C end
		epos=len(hpos)-1
		try:
			while hpos[epos]=='?': epos-=1
		except IndexError:
			hpos.extend(['?', '?'])
		else:
			hep = itertools.cycle('abcdefg')
			while not next(hep) == hpos[epos]: pass
			eadd = [next(hep) for i in range(len(self.res)-epos-1)]
			hpos = hpos[:epos+1] + eadd

		assert len(hpos) == len(self.res)

		# store assigned heptad positions in the object
		if self.ap:
			hpos = hpos[::-1]
		self.positions_twister = hpos

		# Old approach:
		#angles = gen_expected_crick_angles(P, REP, optimal_ph1)
		#self.positions = ['?'] + [crick_to_pos(c, angles)[1] for c in self.crick[1:-1]] + ['?']

	def calc_crickdev(self, P, REP, optimal_ph1=19.5, smooth=False):

		angles = gen_expected_crick_angles(P, REP, optimal_ph1)
		posnames = [chr(97+i) for i in range(len(angles))]

		g = deque(zip(angles, posnames))

		best_fit=1000000

		for i in range(len(g)):
			unpack_g = np.asarray(list(g))

			heptad_cycle = itertools.cycle(unpack_g[:,0].astype(float))

			measured_circk = [r.crick for r in self.res[1:-1]]
			exp_crick = [next(heptad_cycle) for _ in range(len(measured_circk))]

			if self.ap:
				exp_crick = exp_crick[::-1]

			crdev = calc_crick_ang_dev(measured_circk, exp_crick, 0, len(self.res)-2, force_start_Cr_ang_pos=0)[0]

			fit = np.mean([abs(i) for i in crdev])

			if fit < best_fit:
				best_crdev = crdev
				posname_cycle = itertools.cycle(unpack_g[:,1].astype(str))
				best_posnames = [next(posname_cycle) for _ in range(len(best_crdev)+1)]

				first_name = ord(best_posnames[0])-1
				if first_name<97:
					assert first_name==96
					first_name = posnames[-1]
				else:
					first_name = chr(first_name)

				best_posnames = [first_name] + best_posnames

				if self.ap:
					best_posnames = best_posnames[::-1]
				best_fit = fit

			g.rotate(1)

		if smooth:
			self.crdev = [None]+list(savitzky_golay(best_crdev,5,1))+[None]
		else:
			self.crdev = [None]+best_crdev+[None]

		self.positions_samcc = best_posnames

		for pos, i in enumerate(self.crdev):
			self.res[pos].crdev = i



	def calc_crickdev_old(self, P, REP, optimal_ph1=19.5, smooth=False):
		"""
		Calculates Crick Angle deviation.

		Arguments:
			P (int): bundle periodicity
			REP (int): repeat length
		"""

		angles = gen_expected_crick_angles(P, REP, optimal_ph1)

		# No starting heptad for index=0 (no data!) was defined. Define it based on the Crick angle
		if self.heptad == 'x':

			ref_crick = self.res[1].crick
			ref_crick_pos = 1

			if abs(ref_crick)>=165:
				#print(ref_crick, self.res[2].crick)
				ref_crick = self.res[2].crick
				ref_crick_pos = 2

			start_Cr_ang_pos, name, _ = crick_to_pos(ref_crick, angles)
			if not self.ap:
				hpos = start_Cr_ang_pos-ref_crick_pos
			else:
				hpos = start_Cr_ang_pos+ref_crick_pos

			if hpos<0: hpos=REP+hpos  # was -1 -> +hpos
			if hpos>=REP: hpos=hpos-REP   # was == -> >=  ; 0 -> hpos-REP
		else:
			hpos = ord(self.heptad)-97

		# Define heptad positions for all residues
		heptads = ""
		h = hpos
		for _ in range(len(self.res)):
			heptads += chr(h+97)
			if self.ap:
				h-=1
				if h<0: h = REP-1
			else:
				h+=1
				if h==REP: h = 0

		self.heptads = list(heptads)

		hep2angle = dict([(chr(l+97), a) for l,a in enumerate(angles)])

		if self.ap:
			hpos-=1
		else:
			hpos+=1

		if hpos<0:
			hpos = REP-1
		elif hpos==REP:
			hpos = 0

		heptad = chr(hpos+97)

		exp_ahelix_crick = gen_expected_crick_angles(P, REP, hep2angle[heptad], ap=self.ap)

		crdev = calc_crick_ang_dev([r.crick for r in self.res[1:-1]], exp_ahelix_crick, 0, len(self.res)-2,\
										   force_start_Cr_ang_pos=0)[0]


		if smooth:
			self.crdev = [None]+list(savitzky_golay(crdev,5,1))+[None]
		else:
			self.crdev = [None]+crdev+[None]

		for pos, i in enumerate(self.crdev):
			self.res[pos].crdev = i

	def calc_axialshift(self):
		"""C = prevnewC - newC # os CC (nowo wyliczone wartosci z helis, ktorych osie sa przesuniete)
        O = positions[pos]['alphaaxis'][chainnr] - newC # nowa os helisy
        Omove = vector_to_axis(C, O) - O
		"""

		for pos, r in enumerate(self.res[1:-1]):
			realpos=pos+1


			if realpos==1:
				pre = self.res[realpos+1].C
				C = pre - r.C
			else:
				pre = self.res[realpos-1].C
				C = pre - r.C

			O = r.O - r.C

			preOmove = vector_to_axis(C, O)
			Omove = preOmove - O

			temp_newO = r.O + Omove
			temp_distance = (r.O - pre).norm() - (temp_newO - pre).norm()

			r.shift = Omove.norm() * np.sign(temp_distance)
			if realpos==1:
				r.shift*=-1

		self.res[0].shift = None
		self.res[-1].shift = None

		self.shift = [i.shift for i in self.res]



from distutils.core import setup

install_reqs = ['biopython>=1.76', 'networkx==2.2', 'seaborn>=0.10.0']

setup(
  name = 'samcc-turbo',
  packages = ['samcc'],
  version = '0.0.2',
  license='MIT',
  description = 'Software for automatic detection and measurement of coiled coils in PDB structures.',
  author = 'kszczepaniak',
  author_email = 'k.szczepaniak@cent.uw.edu.pl',
  url = 'https://github.com/labstructbioinf/samcc_turbo',
  download_url = 'https://github.com/labstructbioinf/samcc_turbo/archive/v0.0.2.tar.gz',
  keywords = ['bioinformatics', 'protein structure', 'coiled-coil'],
  install_requires=install_reqs,
  classifiers=[
    'Development Status :: 5 - Production/Stable',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    # 'Intended Audience :: Developers',      # Define that your audience are developers
    # 'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    # 'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    # 'Programming Language :: Python :: 3.4',
    # 'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.7',
  ],
)



class layerClass():
	"""Class describing single layer of the bundle"""

	def __init__(self, residues):
		self.res = residues

	def __str__(self):
		return " ".join(["%s:%s" % (r.chain_name, r.res.id[1]) for r in self.res])

	def get_layer_CA(self):
		return [ [res.Ca[0], res.Ca[1], res.Ca[2] ] for res in self.res ]

	def get_layer_axis(self):
		return [ [res.O[0], res.O[1], res.O[2] ] for res in self.res if res.O != None ]

	def get_layer_axis_temp(self):
		return [ type(res.O) for res in self.res ]
